#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import tkinter as tk

from pathlib import Path, PureWindowsPath
from sys import platform
from traceback import format_exc

from binilla.widgets.binilla_widget import BinillaWidget

# inject this default color
BinillaWidget.active_tags_directory_color = '#%02x%02x%02x' % (40, 170, 80)


class DirectoryFrame(BinillaWidget, tk.Frame):
    app_root = None

    def __init__(self, master, *args, **kwargs):
        kwargs.setdefault('app_root', master)
        self.app_root = kwargs.pop('app_root')

        kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)
        BinillaWidget.__init__(self)
        tk.Frame.__init__(self, master, *args, **kwargs)

        self.hierarchy_frame = HierarchyFrame(self, app_root=self.app_root)

        self.hierarchy_frame.pack(fill='both', expand=True)
        self.apply_style()

    def set_root_dir(self, root_dir):
        self.hierarchy_frame.set_root_dir(root_dir)

    def add_root_dir(self, root_dir):
        self.hierarchy_frame.add_root_dir(root_dir)

    def del_root_dir(self, root_dir):
        self.hierarchy_frame.del_root_dir(root_dir)

    def highlight_tags_dir(self, root_dir):
        self.hierarchy_frame.highlight_tags_dir(root_dir)

    def apply_style(self, seen=None):
        self.hierarchy_frame.apply_style(seen)


class HierarchyFrame(BinillaWidget, tk.Frame):
    tags_dir = ''
    app_root = None
    tags_dir_items = ()
    active_tags_dir = ""

    def __init__(self, master, *args, **kwargs):
        kwargs.update(bg=self.default_bg_color, bd=self.listbox_depth,
            relief='sunken', highlightthickness=0)
        kwargs.setdefault('app_root', master)

        select_mode = kwargs.pop('select_mode', 'browse')

        self.app_root = kwargs.pop('app_root')
        BinillaWidget.__init__(self)
        tk.Frame.__init__(self, master, *args, **kwargs)

        self.tags_tree_frame = tk.Frame(self, highlightthickness=0)

        self.tags_tree = tk.ttk.Treeview(
            self.tags_tree_frame, selectmode=select_mode, padding=(0, 0))
        self.scrollbar_y = tk.Scrollbar(
            self.tags_tree_frame, orient='vertical',
            command=self.tags_tree.yview)
        self.tags_tree.config(yscrollcommand=self.scrollbar_y.set)

        self.tags_tree.bind('<<TreeviewOpen>>', self.open_selected)
        self.tags_tree.bind('<<TreeviewClose>>', self.close_selected)
        self.tags_tree.bind('<Double-Button-1>', self.activate_item)
        self.tags_tree.bind('<Return>', self.activate_item)

        self.tags_tree_frame.pack(fill='both', side='left', expand=True)

        # pack in this order so scrollbars aren't shrunk when resizing
        self.scrollbar_y.pack(side='right', fill='y')
        self.tags_tree.pack(side='right', fill='both', expand=True)

        self.reload()
        self.apply_style()

    def apply_style(self, seen=None):
        BinillaWidget.apply_style(self, seen)
        self.tags_tree_frame.config(bg=self.default_bg_color)

        dir_tree = self.tags_tree
        self.highlight_tags_dir()

    def reload(self):
        dir_tree = self.tags_tree
        self.tags_dir = self.app_root.tags_dir
        if not dir_tree['columns']:
            dir_tree['columns'] = ('size', )
            dir_tree.heading("#0", text='path')
            dir_tree.heading("size", text='info')
            dir_tree.column("#0", minwidth=100, width=100)
            dir_tree.column("size", minwidth=100, width=100, stretch=False)

        for tags_dir in self.tags_dir_items:
            dir_tree.delete(tags_dir)

        self.tags_dir_items = []

        for tags_dir in self.app_root.tags_dirs:
            self.add_root_dir(tags_dir)

    def set_root_dir(self, root_dir):
        dir_tree = self.tags_tree
        curr_root_dir = self.app_root.tags_dir

        tags_dir_index = dir_tree.index(curr_root_dir)
        dir_tree.delete(curr_root_dir)
        self.insert_root_dir(root_dir)

    def add_root_dir(self, root_dir):
        '''
        Alias for insert_root_dir.
        '''
        self.insert_root_dir(root_dir)

    def insert_root_dir(self, root_dir, index='end'):
        '''
        Add new top tag directory to the HierarchyFrame.
        '''
        # Turn "/home/currentuser/" into "~/"
        text = str(root_dir)
        if "linux" in platform:
            home = str(Path.home())
            if text.startswith(home):
                text = "~/" + text[1+len(home):]

        iid = self.tags_tree.insert(
            '', index, iid=root_dir, text=text,
            tags=(root_dir, 'tagdir',))
        self.tags_dir_items.append(iid)
        self.destroy_subitems(iid)

    def del_root_dir(self, root_dir):
        self.tags_tree.delete(root_dir)

    def destroy_subitems(self, directory):
        '''
        Destroys all the given items subitems and creates an empty
        subitem so as to give the item the appearance of being expandable.
        '''
        dir_tree = self.tags_tree

        for child in dir_tree.get_children(directory):
            dir_tree.delete(child)

        # add an empty node to make an "expand" button appear
        dir_tree.insert(directory, 'end')

    def generate_subitems(self, directory):
        dir_tree = self.tags_tree

        directory = str(directory)
        for root, subdirs, files in os.walk(directory):
            for subdir in sorted(subdirs, key=str.casefold):
                folderpath = os.path.join(root, subdir)

                dir_info_str = "%s items" % len(list(os.scandir(folderpath)))

                dir_tree.insert(
                    directory, 'end', text=subdir,
                    iid=folderpath, tags=('item',),
                    values=(dir_info_str, ))

                # loop over each of the new items, give them
                # at least one item so they can be expanded.
                self.destroy_subitems(folderpath)

            for file in sorted(files, key=str.casefold):
                fullpath = os.path.join(root, file)
                try:
                    filesize = os.stat(fullpath).st_size
                    if filesize < 1024:
                        filesize = "%d Bytes" % (filesize)
                    elif filesize < 1024**2:
                        filesize = "%.2f KiB" % (filesize/1024)
                    else:
                        filesize = "%.2f MiB" % (filesize/1024**2)
                except Exception:
                    filesize = 'ERROR'
                dir_tree.insert(directory, 'end', text=file,
                                iid=fullpath, tags=('item',),
                                values=(filesize, ))

            # just do the toplevel of the hierarchy
            break

    def get_item_tags_dir(self, iid):
        '''Returns the tags directory of the given item'''
        dir_tree = self.tags_tree
        prev_parent = iid
        parent = dir_tree.parent(prev_parent)

        while parent:
            prev_parent = parent
            parent = dir_tree.parent(prev_parent)

        return prev_parent

    def open_selected(self, e=None):
        dir_tree = self.tags_tree
        tag_path = dir_tree.focus()
        for child in dir_tree.get_children(tag_path):
            dir_tree.delete(child)

        if tag_path:
            self.generate_subitems(tag_path)

    def close_selected(self, e=None):
        dir_tree = self.tags_tree
        tag_path = dir_tree.focus()
        if tag_path is None:
            return

        if Path(tag_path).is_dir():
            self.destroy_subitems(tag_path)

    def highlight_tags_dir(self, tags_dir=None):
        app = self.app_root
        dir_tree = self.tags_tree
        if tags_dir is None:
              tags_dir = self.app_root.tags_dir

        tags_dir = Path(tags_dir)
        for td in app.tags_dirs:
            if td == tags_dir:
                dir_tree.tag_configure(
                    str(td),
                    background=self.active_tags_directory_color,
                    foreground=self.text_highlighted_color)
                self.active_tags_dir = td
            else:
                dir_tree.tag_configure(
                    str(td),
                    background=self.entry_normal_color,
                    foreground=self.text_normal_color)

    def activate_item(self, e=None):
        dir_tree = self.tags_tree
        tag_path = dir_tree.focus()
        if tag_path is None:
            return

        try:
            app = self.app_root
            tags_dir = Path(self.get_item_tags_dir(tag_path))
            tags_dir_index = self.app_root.get_tags_dir_index(
                self.get_item_tags_dir(tag_path))
            if tags_dir_index is None:
                print("'%s' is not a registered tags directory." % tags_dir)
                return

            self.highlight_tags_dir(tags_dir)
            app.switch_tags_dir(index=tags_dir_index)
        except Exception:
            print(format_exc())

        tag_path = Path(tag_path)
        if tag_path.is_dir():
            app.last_load_dir = tag_path
            return

        try:
            app.load_tags(filepaths=tag_path)
        except Exception:
            print(format_exc())


class DependencyFrame(HierarchyFrame):
    root_tag_path = ''
    root_tag_text = None
    _initialized = False
    handler = None

    def __init__(self, master, *args, **kwargs):
        HierarchyFrame.__init__(self, master, *args, **kwargs)
        self.handler = self.app_root.handler
        self._initialized = True

    def apply_style(self, seen=None):
        HierarchyFrame.apply_style(self, seen)
        self.tags_tree.tag_configure(
            'badref', foreground=self.invalid_path_color,
            background=self.default_bg_color)

    def get_item_tags_dir(*args, **kwargs): pass

    def highlight_tags_dir(*args, **kwargs): pass

    def reload(self):
        dir_tree = self.tags_tree
        if self.handler is not None:
            self.tags_dir = self.handler.tagsdir
        else:
            self.tags_dir = Path("")

        if not dir_tree['columns']:
            dir_tree["columns"]=("dependency", )
            dir_tree.heading("#0", text='Filepath')
            dir_tree.heading("dependency", text='Dependency path')

        if not self._initialized:
            return

        for item in dir_tree.get_children():
            try: dir_tree.delete(item)
            except Exception: pass

        root = self.root_tag_path
        text = self.root_tag_text
        if text is None:
            text = root

        iid = self.tags_tree.insert(
            '', 'end', iid=self.root_tag_path, text=text,
            tags=(root, 'item'), values=('', root))
        self.destroy_subitems(iid)

    def get_dependencies(self, tag_path):
        tag = self.master.get_tag(tag_path)
        if tag is None:
            print(("Unable to load '%s'.\n" % tag_path) +
                  "    You may need to change the tag set to load this tag.")
            return ()

        handler = self.handler
        d_id = tag.def_id
        dependency_cache = handler.tag_ref_cache.get(d_id)

        if not dependency_cache:
            return ()

        dependencies = []

        for block in handler.get_nodes_by_paths(dependency_cache, tag.data):
            # if the node's filepath is empty, just skip it
            if not block.filepath:
                continue
            dependencies.append(block)
        return dependencies

    def destroy_subitems(self, iid):
        '''
        Destroys all the given items subitems and creates an empty
        subitem so as to give the item the appearance of being expandable.
        '''
        dir_tree = self.tags_tree

        for child in dir_tree.get_children(iid):
            dir_tree.delete(child)

        # add an empty node to make an "expand" button appear
        tag_path = Path(dir_tree.item(iid)['values'][-1])
        if not tag_path.is_file():
            dir_tree.item(iid, tags=('badref', 'item'))
        elif self.get_dependencies(tag_path):
            dir_tree.insert(iid, 'end')

    def close_selected(self, e=None):
        dir_tree = self.tags_tree
        iid = dir_tree.focus()
        if iid:
            self.destroy_subitems(iid)

    def generate_subitems(self, parent_iid):
        tags_dir = self.handler.tagsdir
        dir_tree = self.tags_tree
        parent_tag_path = Path(dir_tree.item(parent_iid)['values'][-1])

        if not parent_tag_path.is_file():
            return

        for tag_ref_block in self.get_dependencies(parent_tag_path):
            filepath = Path(PureWindowsPath(tag_ref_block.filepath))
            try:
                ext = '.' + tag_ref_block.tag_class.enum_name
                if (self.handler.treat_mode_as_mod2 and ext == '.model' and
                not Path(tags_dir, str(filepath) + ".model").is_file()):
                    ext = '.gbxmodel'
            except Exception:
                ext = ''
            tag_path = str(filepath) + ext

            dependency_name = tag_ref_block.NAME
            last_block = tag_ref_block
            parent = last_block.parent
            while parent is not None and hasattr(parent, 'NAME'):
                name = parent.NAME
                f_type = parent.TYPE
                if f_type.is_array:
                    index = parent.index(last_block)
                    dependency_name = '[%s].%s' % (index, dependency_name)
                elif name not in ('tagdata', 'data'):
                    if not last_block.TYPE.is_array:
                        name += '.'
                    dependency_name = name + dependency_name
                last_block = parent
                parent = last_block.parent

            # slice off the extension and the period
            dependency_name = dependency_name.split('.', 1)[-1]

            iid = dir_tree.insert(
                parent_iid, 'end', text=tag_path, tags=('item',),
                values=(dependency_name, Path(tags_dir, tag_path)))

            self.destroy_subitems(iid)

    def activate_item(self, e=None):
        dir_tree = self.tags_tree
        active = dir_tree.focus()
        if active is None:
            return

        tag_path = Path(dir_tree.item(active)['values'][-1])
        try:
            app = self.app_root
            tags_dir = self.get_item_tags_dir(tag_path)
            self.highlight_tags_dir(tags_dir)
        except Exception:
            print(format_exc())

        if tag_path.is_dir():
            return

        try:
            app.load_tags(filepaths=tag_path)
        except Exception:
            print(format_exc())



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from traceback import format_exc

from binilla.widgets import field_widgets

__all__ = (
    "HaloScriptTextFrame", "HaloHudMessageTextFrame", "DependencyFrame",
    "FontCharacterDisplayFrame", "FontCharacterFrame", "HaloBitmapDisplayFrame",
    "HaloBitmapDisplayButton", "HaloBitmapTagFrame", "HaloBitmapDisplayBase",
    "Halo2BitmapDisplayButton", "Halo2BitmapTagFrame", "Halo3BitmapDisplayFrame",
    "Halo3BitmapDisplayButton", "Halo3BitmapTagFrame", "HaloColorEntry",
    "HaloUInt32ColorPickerFrame", "MeterImageDisplayFrame", "MeterImageFrame",
    "HaloRawdataFrame", "HaloScriptSourceFrame", "SoundSampleFrame",
    "ReflexiveFrame",
    ) + tuple(field_widgets.__all__)

from binilla.widgets.field_widgets import *

from mozzarilla.widgets.field_widgets.computed_text_frames import \
     HaloScriptTextFrame, HaloHudMessageTextFrame
from mozzarilla.widgets.field_widgets.dependency_frame import DependencyFrame
from mozzarilla.widgets.field_widgets.font_display_frame import \
     FontCharacterDisplayFrame, FontCharacterFrame
from mozzarilla.widgets.field_widgets.halo_1_bitmap_display import \
     HaloBitmapDisplayFrame, HaloBitmapDisplayButton, HaloBitmapTagFrame,\
     HaloBitmapDisplayBase
from mozzarilla.widgets.field_widgets.halo_2_bitmap_display import \
     Halo2BitmapDisplayButton, Halo2BitmapTagFrame
from mozzarilla.widgets.field_widgets.halo_3_bitmap_display import \
     Halo3BitmapDisplayFrame, Halo3BitmapDisplayButton, Halo3BitmapTagFrame
from mozzarilla.widgets.field_widgets.halo_color_picker_frame import \
     HaloUInt32ColorPickerFrame, HaloColorEntry
from mozzarilla.widgets.field_widgets.meter_display_frame import \
     MeterImageDisplayFrame, MeterImageFrame
from mozzarilla.widgets.field_widgets.rawdata_frames import HaloRawdataFrame,\
     HaloScriptSourceFrame, SoundSampleFrame
from mozzarilla.widgets.field_widgets.reflexive_frame import ReflexiveFrame

from mozzarilla.widgets.field_widgets import dynamic_enum_frame



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import tkinter as tk

from traceback import format_exc
from struct import unpack, pack

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askopenfilename

from reclaimer.halo_script.hsc_decompilation import extract_h1_scripts

from mozzarilla import editor_constants as e_c


class SauceRemovalWindow(BinillaWidget, tk.Toplevel):
    app_root = None
    handler = None

    print_interval = 5

    def __init__(self, app_root, *args, **kwargs):
        self.handler = app_root.get_handler("Halo 1 OS v4")
        self.app_root = app_root
        kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)
        tk.Toplevel.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, app_root, *args, **kwargs)

        self.title("Scenario Open Sauce remover")
        self.geometry("400x80+0+0")
        self.resizable(0, 0)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        # make the tkinter variables
        self.scenario_path = tk.StringVar(self)

        # make the frames
        self.scenario_frame = tk.LabelFrame(
            self, text="Scenario to remove sauce from")
        self.button_frame = tk.Frame(self)

        self.begin_button = tk.Button(
            self.button_frame, text='Remove sauce',
            width=20, command=self.remove_sauce)
        self.begin_button.tooltip_string = "It's okay, I like buttered noodles too sometimes."

        self.scenario_entry = tk.Entry(
            self.scenario_frame, textvariable=self.scenario_path)
        self.scenario_browse_button = tk.Button(
            self.scenario_frame, text="Browse", command=self.scenario_browse)


        for w in (self.scenario_entry, ):
            w.pack(padx=(4, 0), pady=2, side='left', expand=True, fill='x')

        for w in (self.scenario_browse_button, ):
            w.pack(padx=(0, 4), pady=2, side='left')

        for w in (self.begin_button, ):
            w.pack(padx=4, pady=2, side='left')

        self.scenario_frame.pack(fill='x', padx=1)
        self.button_frame.pack(fill="y")

        self.transient(app_root)
        self.apply_style()
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def scenario_browse(self):
        dirpath = askopenfilename(
            initialdir=self.scenario_path.get(),
            parent=self, title="Select scenario to remove sauce from")

        if not dirpath:
            return
        self.app_root.last_load_dir = os.path.dirname(dirpath)
        self.scenario_path.set(dirpath)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        tk.Toplevel.destroy(self)

    def remove_sauce(self):
        scenario_path = self.scenario_path.get()
        if not os.path.isfile(scenario_path):
            print("Scenario path does not point to a file.")
            return

        print("Scrubbing sauce from scenario...")
        try:
            scnr_tag = self.handler.build_tag(filepath=scenario_path)
        except KeyError:
            scnr_tag = None

        if scnr_tag is None:
            print("    Could not load scenario.")
            return

        tagdata = scnr_tag.data.tagdata

        # make and report any changes that will damage the integrity of the tag
        if tagdata.project_yellow_definitions.filepath:
            print("    WARNING: Removing project yelo definitions.")
            tagdata.project_yellow_definitions.filepath = ""

        if tagdata.bsp_modifiers.size > 0:
            print("    WARNING: Removing all bsp modifiers.")
            del tagdata.bsp_modifiers.STEPTREE[:]

        if tagdata.structure_bsps.size > 16:
            print("    WARNING: More than 16 bsps referenced. " +
                  "Removing the last %s" % (tagdata.structure_bsps.size - 16))
            del tagdata.structure_bsps.STEPTREE[16: ]

        syntax_data = tagdata.script_syntax_data.data
        too_large = len(tagdata.script_string_data.data) > 262144

        if not too_large and len(syntax_data) >= 56:
            # see if there are too many nodes
            last_node = unpack(">H", syntax_data[46: 48])[0]
            too_large &= last_node > 19001

        if too_large:
            print("    WARNING: Scripts are too large to fit in a " +
                  "normal scenario. Extracting scripts to the same " +
                  "directory as the scenario.")
            extract_h1_scripts(
                tagdata,
                os.path.splitext(scenario_path.replace("\\", "/").split("/")[-1])[0],
                out_dir=os.path.dirname(scenario_path), engine="yelo")
            tagdata.script_syntax_data.data = b''
            tagdata.script_string_data.data = b''
        else:
            new_syntax_data = syntax_data[: 32] + pack(">H", 19001)
            tagdata.script_syntax_data.data = new_syntax_data + syntax_data[34: 380076]

        scnr_tag.serialize(backup=True, calc_pointers=False, temp=False)
        print("    Finished")



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import array
import weakref
import tkinter as tk
import tkinter.ttk as ttk

from traceback import format_exc

from reclaimer.bitmaps.meter_image import meter_image_def

from binilla.widgets.scroll_menu import ScrollMenu
from binilla.widgets.bitmap_display_frame import BitmapDisplayFrame
from binilla.widgets.field_widgets import ContainerFrame, SimpleImageFrame

try:
    import arbytmap
except ImportError:
    arbytmap = None


class MeterImageDisplayFrame(BitmapDisplayFrame):
    def __init__(self, *args, **kwargs):
        BitmapDisplayFrame.__init__(self, *args, **kwargs)
        for w in (self.hsb, self.vsb, self.root_canvas):
            w.pack_forget()

        self.preview_label = tk.Label(self, text="Bitmap preview\t")
        self.image_canvas = tk.Canvas(self, highlightthickness=0)
        self.channel_menu = ScrollMenu(self, menu_width=9, can_scroll=True,
                                       variable=self.channel_index)
        self.save_button = ttk.Button(self, width=11, text="Save as...",
                                      command=self.save_as)

        padx = self.horizontal_padx
        pady = self.horizontal_pady

        self.image_canvas.config(width=1, height=1)
        self.preview_label.pack(side='left', padx=padx, pady=pady, fill='x')
        self.channel_menu.pack(side='left', padx=padx, pady=pady)
        self.save_button.pack(side='left', padx=padx, pady=pady)
        self.image_canvas.pack(side='left', padx=padx, pady=pady,
                               fill='x', expand=True)

        self.apply_style()

    def apply_style(self, seen=None):
        BitmapDisplayFrame.apply_style(self, seen)
        self.image_canvas.config(bg=self.bitmap_canvas_bg_color)


class MeterImageFrame(SimpleImageFrame):
    display_frame_cls = MeterImageDisplayFrame

    def reload(self):
        ContainerFrame.reload(self)
        if self.image_frame and self.node:
            try:
                width = min(640, max(self.node.width, 1))
                height = min(480, max(self.node.height, 1))
                self.image_frame().change_textures(self.get_textures())
            except Exception:
                print(format_exc())
                width = height = 1
            self.image_frame().image_canvas.config(width=width, height=height)

        self.pose_fields()

    def get_textures(self):
        texture_block = []
        if self.node:
            # apparently some meter images tags have fucked endianness
            # on the dimensions and position, so these NEED to be capped
            width = min(640, max(self.node.width, 1))
            height = min(480, max(self.node.height, 1))
            tex_info = dict(width=width, height=height,
                            format=arbytmap.FORMAT_A8R8G8B8)
            pixels = bytearray(width * height * 4)
            for line in meter_image_def.build(rawdata=self.node.meter_data.data):
                off = (line.x_pos + line.y_pos * width) * 4
                pixels[off: off + line.width * 4] = line.line_data

            # need it packed otherwise channel swapping wont occur
            texture_block.append([[array.array("I", pixels)], tex_info])

        return texture_block



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

try:
    from .converter_base import ConverterBase
except (ImportError, SystemError):
    from converter_base import ConverterBase

from pathlib import Path, PureWindowsPath
import threadsafe_tkinter as tk

from copy import deepcopy
from struct import Struct as PyStruct
from traceback import format_exc

from reclaimer.model.jms import ( JmsNode, JmsMaterial, JmsMarker,
     JmsVertex, JmsTriangle, JmsModel, MergedJmsModel, )
from reclaimer.model.jms.util import edge_loop_to_tris
from reclaimer.model.model_compilation import compile_gbxmodel
from reclaimer.util.matrices import euler_to_quaternion, Ray
from reclaimer.util.geometry import planes_to_verts_and_edge_loops
from reclaimer.hek.defs.sbsp import sbsp_def
from reclaimer.hek.defs.mod2 import mod2_def

from supyr_struct.defs.block_def import BlockDef

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk


def planes_to_verts_and_tris(planes, center, region=0, mat_id=0,
                             make_fans=False, round_adjust=0.000001):
    raw_verts, edge_loops = planes_to_verts_and_edge_loops(
        planes, center, round_adjust=round_adjust)

    verts = [JmsVertex(0, v[0]*100, v[1]*100, v[2]*100, tex_v=1.0) for v in raw_verts]
    tris = []
    # Calculate verts and triangles from the raw vert positions and edge loops
    for edge_loop in edge_loops:
        tris.extend(edge_loop_to_tris(edge_loop, region, mat_id, 0, make_fans))

    return verts, tris


def get_bsp_surface_edge_loops(bsp, ignore_flags=False):
    surfaces = bsp.surfaces.STEPTREE
    edges = bsp.edges.STEPTREE

    edge_loops = {}
    # loop over each surface in the collision.
    # NOTE: These are polygonal, not just triangular
    for s_i in range(len(surfaces)):
        surface = surfaces[s_i]
        flags = surface.flags
        e_i = surface.first_edge
        if ignore_flags:
            key = (surface.material, )
        else:
            key = (surface.material, flags.two_sided, flags.invisible,
                   flags.climbable, flags.breakable)

        surface_edges = set()
        vert_indices = []
        # loop over each edge in the surface and concatenate
        # the verts that make up the outline until we run out
        while e_i not in surface_edges:
            surface_edges.add(e_i)
            edge = edges[e_i]
            if edge[4] == s_i:
                e_i = edge[2]
                vert_indices.append(edge[0])
            else:
                e_i = edge[3]
                vert_indices.append(edge[1])

        edge_loops.setdefault(key, []).append(vert_indices)

    return edge_loops


def make_bsp_jms_verts(bsp, transform=None):
    verts = [JmsVertex(0, v[0]*100, v[1]*100, v[2]*100, tex_v=1.0)
             for v in bsp.vertices.STEPTREE]
    return verts


def make_marker_jms_model(sbsp_markers, nodes):
    markers = []
    for m in sbsp_markers:
        markers.append(
            JmsMarker(m.name, "bsp", 0, 0, m.rotation.i,
                      m.rotation.j, m.rotation.k, m.rotation.w,
                      m.position.x*100, m.position.y*100, m.position.z*100)
            )

    return JmsModel("bsp", 0, nodes, [], markers, ("markers", ))


def make_lens_flare_jms_model(lens_flare_markers, lens_flares, nodes):
    markers = []
    for m in lens_flare_markers:
        i, j, k, w = euler_to_quaternion(
            m.direction.i/128, m.direction.j/128, m.direction.k/128)
        lens_flare_name = lens_flares[m.lens_flare_index].shader.filepath

        markers.append(
            JmsMarker(lens_flare_name.split("\\")[-1], "bsp", 0, 0, i, j, k, w,
                      m.position.x*100, m.position.y*100, m.position.z*100)
            )

    return JmsModel("bsp", 0, nodes, [], markers, ("lens_flares", ))


def make_mirror_jms_models(clusters, nodes, make_fans=True):
    jms_models = []
    mirrors = []

    for cluster in clusters:
        mirrors.extend(cluster.mirrors.STEPTREE)

    mirror_index = 0
    for mirror in mirrors:
        tris = edge_loop_to_tris(
            len(mirror.vertices.STEPTREE), make_fan=make_fans)
        verts = [
            JmsVertex(0, vert[0] * 100, vert[1] * 100, vert[2] * 100, tex_v=1.0)
            for vert in mirror.vertices.STEPTREE
            ]

        jms_models.append(
            JmsModel("bsp", 0, nodes,
                     [JmsMaterial(mirror.shader.filepath.split("\\")[-1])],
                     (), ("mirror_%s" % mirror_index, ), verts, tris))

        mirror_index += 1

    return jms_models


def make_fog_plane_jms_models(fog_planes, nodes, make_fans=True, optimize=False):
    jms_models = []
    materials = [JmsMaterial("+unused$", "<none>", "+unused$")]

    plane_index = 0
    for fog_plane in fog_planes:
        tris = edge_loop_to_tris(
            len(fog_plane.vertices.STEPTREE), make_fan=make_fans)
        verts = [
            JmsVertex(0, vert[0] * 100, vert[1] * 100, vert[2] * 100, tex_v=1.0)
            for vert in fog_plane.vertices.STEPTREE
            ]


        jms_model = JmsModel(
            "bsp", 0, nodes, materials, (),
            ("fog_plane_%s" % plane_index, ), verts, tris)

        if optimize:
            jms_model.optimize_geometry(True)

        jms_models.append(jms_model)
        plane_index += 1

    return jms_models


def make_cluster_portal_jms_models(planes, clusters, cluster_portals, nodes,
                                   make_fans=True, optimize=False):
    jms_models = []
    materials = [
        JmsMaterial("+portal", "<none>", "+portal"),
        JmsMaterial("+portal&", "<none>", "+portal&")
        ]
    materials[1].properties = ""

    cluster_index = 0
    portals_seen = set()
    verts = []
    tris = []
    for cluster in clusters:
        for portal_index in cluster.portals.STEPTREE:
            if portal_index[0] in portals_seen:
                continue

            portals_seen.add(portal_index[0])
            portal = cluster_portals[portal_index[0]]
            shader = 1 if portal.flags.ai_cant_hear_through_this else 0
            portal_plane = planes[portal.plane_index]

            tris.extend(edge_loop_to_tris(
                len(portal.vertices.STEPTREE), mat_id=shader,
                base=len(verts), make_fan=make_fans)
                )
            verts.extend(
                JmsVertex(0, vert[0] * 100, vert[1] * 100, vert[2] * 100, tex_v=1.0)
                for vert in portal.vertices.STEPTREE
                )

    jms_model = JmsModel(
        "bsp", 0, nodes, materials, (),
        ("cluster_portals", ), verts, tris)

    if optimize:
        jms_model.optimize_geometry(True)

    jms_models.append(jms_model)
    return jms_models


def make_weather_polyhedra_jms_models(polyhedras, nodes, make_fans=True,
                                      tolerance=0.0000001):
    jms_models = []
    materials = [JmsMaterial("+weatherpoly", "<none>", "+weatherpoly")]

    polyhedra_index = 0
    for polyhedra in polyhedras:
        verts, tris = planes_to_verts_and_tris(
            polyhedra.planes.STEPTREE, polyhedra.bounding_sphere_center,
            make_fans=make_fans,
            round_adjust=Ray(polyhedra.bounding_sphere_center).mag * tolerance)

        jms_models.append(JmsModel(
            "bsp", 0, nodes, materials, (),
            ("weather_polyhedra_%s" % polyhedra_index, ), verts, tris))

        polyhedra_index += 1

    return jms_models


def make_bsp_coll_jms_models(bsps, materials, nodes, node_transforms=(),
                             ignore_flags=False, make_fans=True):
    jms_models = []
    bsp_index = 0
    for bsp in bsps:
        coll_edge_loops = get_bsp_surface_edge_loops(bsp, ignore_flags)
        node_transform = node_transforms[bsp_index] if node_transforms else None

        coll_materials = []
        mat_info_to_mat_id = {}
        # create materials from the provided materials and the
        # info on the collision properties of each surface.
        for mat_info in coll_edge_loops:
            src_material = materials[mat_info[0]]
            material = JmsMaterial(src_material.name)
            if not ignore_flags:
                if len(mat_info) > 1: material.double_sided = mat_info[1]
                if len(mat_info) > 2: material.large_collideable = mat_info[2]
                if len(mat_info) > 3: material.ladder = mat_info[3]
                if len(mat_info) > 4: material.breakable = mat_info[4]
                material.collision_only = not material.large_collideable
                material.double_sided &= not material.large_collideable
                material.name = material.name + material.properties
                material.shader_path = material.shader_path + material.properties
                material.properties = ""

            mat_info_to_mat_id[mat_info] = len(coll_materials)
            coll_materials.append(material)

        verts = make_bsp_jms_verts(bsp, node_transform)

        tri_count = 0
        # figure out how many triangles we'll be creating
        for mat_info in coll_edge_loops:
            for edge_loop in coll_edge_loops[mat_info]:
                tri_count += len(edge_loop) - 2

        tri_index = 0
        tris = [None] * tri_count
        # create triangles from the edge loops
        for mat_info in coll_edge_loops:
            mat_id = mat_info_to_mat_id[mat_info]
            for edge_loop in coll_edge_loops[mat_info]:
                loop_tris = edge_loop_to_tris(
                    edge_loop, mat_id=mat_id, make_fan=make_fans)
                tris[tri_index: tri_index + len(loop_tris)] = loop_tris
                tri_index += len(loop_tris)

        jms_models.append(
            JmsModel("bsp", 0, nodes, coll_materials, [],
                     ("collision_%s" % bsp_index, ), verts, tris))
        bsp_index += 1

    return jms_models


def make_bsp_lightmap_jms_models(sbsp_body, base_nodes):
    jms_models = []

    lightmaps = sbsp_body.lightmaps.STEPTREE
    all_tris = sbsp_body.surfaces.STEPTREE

    shader_index_by_mat_name = {}
    shader_mats = []
    for i in range(len(lightmaps)):
        lm_index = lightmaps[i].bitmap_index
        if lm_index not in shader_index_by_mat_name and lm_index >= 0:
            shader_index_by_mat_name[lm_index] = len(shader_index_by_mat_name)
            shader_mats.append(JmsMaterial("lightmap_%s" % lm_index))

    uncomp_vert_xyz_unpacker = PyStruct("<3f").unpack_from
    uncomp_vert_ijkuv_unpacker = PyStruct("<5f").unpack_from

    for lightmap in lightmaps:
        verts = []
        tris = []
        mat_index = shader_index_by_mat_name.get(lightmap.bitmap_index, -1)
        if mat_index < 0:
            continue

        for material in lightmap.materials.STEPTREE:
            v_base = len(verts)
            tris.extend(
                JmsTriangle(
                    0, mat_index,
                    tri[0] + v_base, tri[2] + v_base, tri[1] + v_base)
                for tri in all_tris[
                    material.surfaces: material.surfaces + material.surface_count]
                )

            vert_off = 0
            lm_vert_off = 56 * material.vertices_count
            vert_data = material.uncompressed_vertices.data
            for i in range(material.lightmap_vertices_count):
                x, y, z = uncomp_vert_xyz_unpacker(vert_data, vert_off)
                i, j, k, u, v = uncomp_vert_ijkuv_unpacker(vert_data, lm_vert_off)
                vert_off += 56
                lm_vert_off += 20
                verts.append(
                    JmsVertex(0, x * 100, y * 100, z * 100,
                              i, j, k, -1, 0, u, 1 - v)
                )

        jms_models.append(
            JmsModel("bsp", 0, base_nodes, shader_mats, [],
                     ("lightmap_%s" % lightmap.bitmap_index, ), verts, tris))

    return jms_models


def make_bsp_renderable_jms_models(sbsp_body, base_nodes):
    jms_models = []

    lightmaps = sbsp_body.lightmaps.STEPTREE
    all_tris = sbsp_body.surfaces.STEPTREE

    shader_index_by_mat_name = {}
    mat_indices_by_mat_name = {}
    shader_mats = []
    for i in range(len(lightmaps)):
        materials = lightmaps[i].materials.STEPTREE
        for j in range(len(materials)):
            material = materials[j]
            mat_name = PureWindowsPath(material.shader.filepath).name.lower()
            mat_name += "!$" if material.flags.fog_plane else "!"

            if mat_name not in mat_indices_by_mat_name:
                shader_index_by_mat_name[mat_name] = len(shader_mats)
                shader_mats.append(JmsMaterial(mat_name))
                mat_indices_by_mat_name[mat_name] = []
                shader_mats[-1].shader_path = (shader_mats[-1].shader_path +
                                               shader_mats[-1].properties)
                shader_mats[-1].properties = ""

            mat_indices_by_mat_name[mat_name].append((i, j))

    uncomp_vert_unpacker = PyStruct("<14f").unpack_from
    for mat_name in sorted(mat_indices_by_mat_name):
        verts = []
        tris = []
        for i, j in mat_indices_by_mat_name[mat_name]:
            material = lightmaps[i].materials.STEPTREE[j]

            mat_index = shader_index_by_mat_name.get(mat_name)
            if mat_index is None:
                continue

            vert_data = material.uncompressed_vertices.data
            v_base = len(verts)

            tris.extend(
                JmsTriangle(
                    0, mat_index,
                    tri[0] + v_base, tri[2] + v_base, tri[1] + v_base)
                for tri in all_tris[
                    material.surfaces: material.surfaces + material.surface_count]
                )

            for i in range(0, material.vertices_count * 56, 56):
                x, y, z, ni, nj, nk, bi, bj, bk, ti, tj, tk, u, v =\
                   uncomp_vert_unpacker(vert_data, i)
                verts.append(
                    JmsVertex(0, x * 100, y * 100, z * 100,
                              ni, nj, nk, -1, 0, u, 1 - v, 0,
                              bi, bj, bk, ti, tj, tk)
                )

        jms_models.append(
            JmsModel("bsp", 0, base_nodes, shader_mats, [],
                     ("renderable", ), verts, tris))

    return jms_models


def sbsp_to_mod2(
        sbsp_path, include_lens_flares=True, include_markers=True,
        include_weather_polyhedra=True, include_fog_planes=True,
        include_portals=True, include_collision=True, include_renderable=True,
        include_mirrors=True, include_lightmaps=True, fan_weather_polyhedra=True,
        fan_fog_planes=True,  fan_portals=True, fan_collision=True,
        fan_mirrors=True, optimize_fog_planes=False, optimize_portals=False,
        weather_polyhedra_tolerance=0.0000001):

    print("    Loading sbsp tag...")
    sbsp_tag = sbsp_def.build(filepath=sbsp_path)
    mod2_tag = mod2_def.build()

    sbsp_body = sbsp_tag.data.tagdata
    coll_mats = [JmsMaterial(mat.shader.filepath.split("\\")[-1])
                 for mat in sbsp_body.collision_materials.STEPTREE]

    base_nodes = [JmsNode("frame")]
    jms_models = []

    if include_markers:
        print("    Converting markers...")
        try:
            jms_models.append(make_marker_jms_model(
                sbsp_body.markers.STEPTREE, base_nodes))
        except Exception:
            print(format_exc())
            print("    Could not convert markers")

    if include_lens_flares:
        print("    Converting lens flares...")
        try:
            jms_models.append(make_lens_flare_jms_model(
                sbsp_body.lens_flare_markers.STEPTREE,
                sbsp_body.lens_flares.STEPTREE, base_nodes))
        except Exception:
            print(format_exc())
            print("    Could not convert lens flares")

    if include_fog_planes:
        print("    Converting fog planes...")
        try:
            jms_models.extend(make_fog_plane_jms_models(
                sbsp_body.fog_planes.STEPTREE, base_nodes,
                fan_fog_planes, optimize_fog_planes))
        except Exception:
            print(format_exc())
            print("    Could not convert fog planes")

    if include_mirrors:
        print("    Converting mirrors...")
        try:
            jms_models.extend(make_mirror_jms_models(
                sbsp_body.clusters.STEPTREE, base_nodes, fan_mirrors))
        except Exception:
            print(format_exc())
            print("    Could not convert mirrors")

    if include_portals and sbsp_body.collision_bsp.STEPTREE:
        print("    Converting portals...")
        try:
            jms_models.extend(make_cluster_portal_jms_models(
                sbsp_body.collision_bsp.STEPTREE[0].planes.STEPTREE,
                sbsp_body.clusters.STEPTREE, sbsp_body.cluster_portals.STEPTREE,
                base_nodes, fan_portals, optimize_portals))
        except Exception:
            print(format_exc())
            print("    Could not convert portals")

    if include_weather_polyhedra:
        print("    Converting weather polyhedra...")
        try:
            jms_models.extend(make_weather_polyhedra_jms_models(
                sbsp_body.weather_polyhedras.STEPTREE, base_nodes,
                fan_weather_polyhedra, weather_polyhedra_tolerance))
        except Exception:
            print(format_exc())
            print("    Could not convert weather polyhedra")

    if include_collision:
        print("    Converting collision...")
        try:
            jms_models.extend(make_bsp_coll_jms_models(
                sbsp_body.collision_bsp.STEPTREE, coll_mats, base_nodes,
                None, False, fan_collision))
        except Exception:
            print(format_exc())
            print("    Could not convert collision")

    if include_renderable:
        print("    Converting renderable...")
        try:
            jms_models.extend(
                make_bsp_renderable_jms_models(sbsp_body, base_nodes))
        except Exception:
            print(format_exc())
            print("    Could not convert renderable")

    if include_lightmaps:
        print("    Converting lightmaps...")
        try:
            jms_models.extend(
                make_bsp_lightmap_jms_models(sbsp_body, base_nodes))
        except Exception:
            print(format_exc())
            print("    Could not convert lightmaps")

    print("    Compiling gbxmodel...")
    mod2_tag.filepath = str(Path(sbsp_path).with_suffix('')) + "_SBSP.gbxmodel"
    compile_gbxmodel(mod2_tag, MergedJmsModel(*jms_models), True)
    return mod2_tag


class SbspConverter(ConverterBase, window_base_class):
    src_ext = "scenario_structure_bsp"
    dst_ext = "gbxmodel"

    weather_tolerance = 0.0000001
    min_weather_tolerance = 0.000000000000001

    def __init__(self, app_root, *args, **kwargs):
        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)

        window_base_class.__init__(self, app_root, *args, **kwargs)
        ConverterBase.__init__(self, app_root, *args, **kwargs)
        self.setup_window(*args, **kwargs)

    def setup_window(self, *args, **kwargs):
        ConverterBase.setup_window(self, *args, **kwargs)

        self.include_weather_polyhedra = tk.IntVar(self, 1)
        self.include_fog_planes = tk.IntVar(self, 1)
        self.include_portals = tk.IntVar(self, 1)
        self.include_collision = tk.IntVar(self, 1)
        self.include_renderable = tk.IntVar(self, 1)
        self.include_mirrors = tk.IntVar(self, 0)
        self.include_lightmaps = tk.IntVar(self, 0)

        self.include_markers = tk.IntVar(self, 1)
        self.include_lens_flares = tk.IntVar(self, 0)

        self.fan_portals = tk.IntVar(self, 1)
        self.fan_weather_polyhedra = tk.IntVar(self, 1)
        self.fan_fog_planes = tk.IntVar(self, 1)
        self.fan_mirrors = tk.IntVar(self, 1)
        self.fan_collision = tk.IntVar(self, 1)

        self.optimize_portals = tk.IntVar(self, 0)
        self.optimize_fog_planes = tk.IntVar(self, 0)
        self.weather_tolerance_string = tk.StringVar(self, str(self.weather_tolerance))
        self.weather_tolerance_string.trace(
            "w", lambda *a, s=self: s.set_weather_tolerance())

        # make the frames
        self.include_frame = tk.LabelFrame(self, text="Geometry/markers to include")
        self.weather_tolerance_frame = tk.LabelFrame(self, text="Weather polyhedron tolerance")
        self.topology_frame = tk.LabelFrame(self, text="Topology generation")


        # Generate the important frame and its contents
        include_vars = {
            "Weather polyhedra": self.include_weather_polyhedra,
            "Fog planes": self.include_fog_planes, "Portals": self.include_portals,
            "Collidable": self.include_collision, "Renderable": self.include_renderable,
            "Mirrors": self.include_mirrors, "Lightmaps": self.include_lightmaps,
            "Markers": self.include_markers, "Lens flares": self.include_lens_flares}
        self.include_buttons = []
        for text in ("Collidable", "Portals", "Renderable",
                     "Weather polyhedra", "Fog planes", "Markers",
                     "Mirrors", "Lens flares", "Lightmaps"):
            self.include_buttons.append(tk.Checkbutton(
                self.include_frame, variable=include_vars[text], text=text))

        # Generate the topology frame and its contents
        topology_vars = {
            "Weather polyhedra": self.fan_weather_polyhedra,
            "Fog planes": self.fan_fog_planes, "Mirrors": self.fan_mirrors,
            "Portals": self.fan_portals, "Collision": self.fan_collision}
        self.topology_frames = []
        self.topology_labels = []
        self.topology_buttons = []
        for text in ("Portals", "Fog planes", "Weather polyhedra", "Mirrors", "Collision"):
            var = topology_vars[text]
            f = tk.Frame(self.topology_frame)
            name_lbl = tk.Label(f, text=text, width=17, anchor="w")
            fan_cbtn = tk.Checkbutton(
                f, variable=var, text="Triangle fan")
            strip_cbtn = tk.Checkbutton(
                f, variable=var, text="Triangle strip", onvalue=0, offvalue=1)
            self.topology_frames.append(f)
            self.topology_labels.append(name_lbl)
            self.topology_buttons.extend((fan_cbtn, strip_cbtn))
            if text == "Portals":
                self.topology_buttons.append(tk.Checkbutton(
                    f, variable=self.optimize_portals, text="Optimize"))
            elif text == "Fog planes":
                self.topology_buttons.append(tk.Checkbutton(
                    f, variable=self.optimize_fog_planes, text="Optimize"))

        self.weather_tolerance_info = tk.Label(
            self.weather_tolerance_frame, justify='left', anchor="w",
            text=("Due to how weather polyhedrons work, there is no geometry to extract, so it must be generated. \n"
                  "My method for doing this isn't perfect, so sometimes geometry will be missing faces. Adjust this\n"
                  "value to find the sweet spot. NEVER set to 0, and be wary of setting to 0.0001 or higher.\n"
                  "\tNOTE: You will probably need to manually clean up the generated geometry a bit."))
        self.weather_tolerance_spinbox = tk.Spinbox(
            self.weather_tolerance_frame, from_=self.min_weather_tolerance,
            to=100, width=25, increment=self.weather_tolerance,
            textvariable=self.weather_tolerance_string, justify="right")

        self.pack_widgets()
        self.apply_style()

    def pack_widgets(self):
        ConverterBase.pack_widgets(self)

        # pack everything
        for frame in self.topology_frames:
            frame.pack(expand=True, fill='both')

        for label in self.topology_labels:
            label.pack(anchor='w', padx=10, side='left')

        x = y = 0
        for button in self.include_buttons:
            button.grid(row=y, column=x, padx=5, pady=5, sticky="w")
            x += 1
            if x == 5:
                x = 0
                y += 1

        for button in self.topology_buttons:
            button.pack(anchor='w', padx=5, side='left')

        self.weather_tolerance_info.pack(fill='both', expand=True, padx=5, pady=5)
        self.weather_tolerance_spinbox.pack(padx=5, pady=5)

        self.include_frame.pack(expand=True, fill='both')
        self.topology_frame.pack(expand=True, fill='both')
        self.weather_tolerance_frame.pack(expand=True, fill='both')

    def lock_ui(self):
        ConverterBase.lock_ui(self)
        for w in self.include_buttons: w.config(state=tk.DISABLED)
        for w in self.topology_buttons: w.config(state=tk.DISABLED)
        self.weather_tolerance_spinbox.config(state=tk.DISABLED)

    def unlock_ui(self):
        ConverterBase.unlock_ui(self)
        for w in self.include_buttons: w.config(state=tk.NORMAL)
        for w in self.topology_buttons: w.config(state=tk.NORMAL)
        self.weather_tolerance_spinbox.config(state=tk.NORMAL)

    def set_weather_tolerance(self):
        try:
            new_tolerance = float(self.weather_tolerance_string.get())
            if new_tolerance >= self.min_weather_tolerance:
                self.weather_tolerance = new_tolerance
                return

            self.weather_tolerance = self.min_weather_tolerance
        except Exception:
            return

        self.weather_tolerance_string.set(
            str(("%.20f" % self.weather_tolerance)).rstrip("0").rstrip("."))

    def destroy(self):
        ConverterBase.destroy(self)
        window_base_class.destroy(self)

    def convert(self, tag_path=None):
        return sbsp_to_mod2(
            tag_path, self.include_lens_flares.get(),
            self.include_markers.get(), self.include_weather_polyhedra.get(),
            self.include_fog_planes.get(), self.include_portals.get(),
            self.include_collision.get(), self.include_renderable.get(),
            self.include_mirrors.get(), self.include_lightmaps.get(),
            self.fan_weather_polyhedra.get(), self.fan_fog_planes.get(),
            self.fan_portals.get(), self.fan_collision.get(), self.fan_mirrors.get(),
            self.optimize_fog_planes.get(), self.optimize_portals.get(),
            self.weather_tolerance)


if __name__ == "__main__":
    try:
        SbspConverter(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#
from traceback import format_exc

from binilla.widgets.field_widgets.enum_frame import DynamicEnumFrame
from binilla.constants import DYN_NAME_PATH, DYN_I

__all__ = ("DynamicEnumFrame", )


# replace the DynamicEnumFrame with one that has a specialized option generator
def halo_dynamic_enum_generate_options(self, opt_index=None):
    desc = self.desc
    options = {0: "-1. NONE"}

    dyn_name_path = desc.get(DYN_NAME_PATH)
    if self.node is None:
        if opt_index is None:
            return options
        return None
    elif not dyn_name_path:
        print("Missing DYN_NAME_PATH path in dynamic enumerator.")
        print(self.parent.get_root().def_id, self.name)
        if opt_index is None:
            return options
        return None

    try:
        p_out, p_in = dyn_name_path.split(DYN_I)

        # We are ALWAYS going to go to the parent, so we need to slice
        if p_out.startswith('..'): p_out = p_out.split('.', 1)[-1]
        array = self.parent.get_neighbor(p_out)

        options_to_generate = range(len(array))
        if opt_index is not None:
            options_to_generate = (
                (opt_index - 1, ) if opt_index - 1 in
                options_to_generate else ())

        for i in options_to_generate:
            name = array[i].get_neighbor(p_in)
            if isinstance(name, list):
                name = repr(name).strip("[").strip("]")
            else:
                name = str(name)

            if p_in.endswith('.filepath'):
                # if it is a dependency filepath
                trimmed_name = name.replace('/', '\\').split('\\')[-1]
                if trimmed_name.strip():
                    name = trimmed_name

            options[i + 1] = '%s. %s' % (i, name)

        last_option_index = len(array)
    except Exception:
        print(format_exc())
        last_option_index = 0

    if opt_index is None:
        self.option_cache = options
        self.options_sane = True
        if self.sel_menu is not None:
            self.sel_menu.options_menu_sane = False
            self.sel_menu.max_index = last_option_index
        return options
    return options.get(opt_index, None)

DynamicEnumFrame.generate_options = halo_dynamic_enum_generate_options



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = (
    "tools", "tag_converters", "tag_window"
    )

from mozzarilla.windows import tools, tag_converters, tag_window



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import threadsafe_tkinter as tk
import zlib

from time import time
from threading import Thread
from struct import unpack, pack_into
from traceback import format_exc

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askdirectory
from mozzarilla import editor_constants as e_c

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk


class BitmapSourceExtractorWindow(BinillaWidget, window_base_class):
    _running_thread = None
    stop_extraction = False

    def __init__(self, app_root, *args, **kwargs):
        self.app_root = app_root

        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)

        window_base_class.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, app_root, *args, **kwargs)

        self.title("Halo 1 & 2 bitmap source extractor")
        self.resizable(0, 0)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        self.tags_dir = tk.StringVar(self)
        self.data_dir = tk.StringVar(self)
        self.tags_dir.set(e_c.WORKING_DIR.joinpath('tags'))
        self.data_dir.set(e_c.WORKING_DIR.joinpath('data'))

        # make the frames
        self.tags_dir_frame = tk.LabelFrame(self, text="Tags directory")
        self.data_dir_frame = tk.LabelFrame(self, text="Data directory")

        # add the filepath boxes
        self.tags_dir_entry = tk.Entry(
            self.tags_dir_frame, textvariable=self.tags_dir)
        self.tags_dir_entry.config(width=55, state=tk.DISABLED)
        self.data_dir_entry = tk.Entry(
            self.data_dir_frame, textvariable=self.data_dir)
        self.data_dir_entry.config(width=55, state=tk.DISABLED)

        # add the buttons
        self.extract_btn = tk.Button(
            self, text="Extract source files", width=22,
            command=self.extract)
        self.tags_dir_browse_btn = tk.Button(
            self.tags_dir_frame, text="Browse",
            width=6, command=self.tags_dir_browse)
        self.data_dir_browse_btn = tk.Button(
            self.data_dir_frame, text="Browse",
            width=6, command=self.data_dir_browse)

        # pack everything
        self.tags_dir_entry.pack(expand=True, fill='x', side='left')
        self.data_dir_entry.pack(expand=True, fill='x', side='left')
        self.tags_dir_browse_btn.pack(fill='x', side='left')
        self.data_dir_browse_btn.pack(fill='x', side='left')

        self.tags_dir_frame.pack(expand=True, fill='both')
        self.data_dir_frame.pack(expand=True, fill='both')
        self.extract_btn.pack(fill='both', padx=5, pady=5)

        if self.app_root is not self and self.app_root:
            self.transient(self.app_root)

        self.apply_style()

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        self.stop_extraction = True
        super(BitmapSourceExtractorWindow, self).destroy()

    def apply_style(self, seen=None):
        super(BitmapSourceExtractorWindow, self).apply_style(seen)
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def tags_dir_browse(self):
        dirpath = askdirectory(initialdir=self.tags_dir.get())
        if dirpath:
            self.tags_dir.set(dirpath)

    def data_dir_browse(self):
        dirpath = askdirectory(initialdir=self.data_dir.get())
        if dirpath:
            self.data_dir.set(dirpath)

    def lock_ui(self):
        for w in (self.tags_dir_browse_btn, self.extract_btn,
                  self.data_dir_browse_btn):
            w.config(state=tk.DISABLED)

    def unlock_ui(self):
        for w in (self.tags_dir_browse_btn, self.extract_btn,
                  self.data_dir_browse_btn):
            w.config(state=tk.NORMAL)

    def thread_wrapper(self, func, *args, **kwargs):
        if self._running_thread is not None:
            return

        try:
            self.lock_ui()
            func(*args, **kwargs)
            self.unlock_ui()
            self._running_thread = None
        except Exception:
            self.unlock_ui()
            self._running_thread = None
            raise

    def extract(self):
        if self._running_thread is None:
            new_thread = Thread(target=self.thread_wrapper,
                                args=(self.do_extract, ))
            new_thread.daemon = True
            new_thread.start()

    def do_extract(self):
        print('Extracting source files...')
        start = time()
        tags_dir = self.tags_dir.get()
        data_dir = self.data_dir.get()

        for root, dirs, files in os.walk(tags_dir):
            for tag_name in files:
                if self.stop_extraction:
                    print("    Conversion cancelled by user.")
                    return
                elif os.path.splitext(tag_name)[-1].lower() != '.bitmap':
                    continue
                tag_path = os.path.join(root, tag_name)

                source_path = data_dir + tag_path.split(tags_dir)[-1]
                source_path = os.path.splitext(source_path)[0] + ".tga"
                source_dir = os.path.dirname(source_path)

                try:
                    with open(tag_path, 'rb') as f:
                        data = f.read()

                    tag_id = data[36:40]
                    engine_id = data[60:64]

                    # make sure this is a bitmap tag
                    if tag_id == b'bitm' and engine_id == b'blam':
                        dims_off = 64+24
                        size_off = 64+28
                        data_off = 64+108
                        end = ">"
                    elif tag_id == b'mtib' and engine_id == b'!MLB':
                        dims_off = 64+16+24
                        size_off = 64+16+28
                        data_off = 64+16
                        # get the size of the bitmap body from the tbfd structure
                        data_off += unpack("<i", data[data_off-4: data_off])[0]
                        end = "<"
                    else:
                        #print("    This file doesnt appear to be a bitmap tag.")
                        continue

                    width, height = unpack(end+"HH", data[dims_off: dims_off+4])
                    comp_size = unpack(end+"i", data[size_off: size_off+4])[0]
                    data = data[data_off: data_off+comp_size]
                except Exception:
                    #print("    Could not load bitmap tag.")
                    continue

                if not len(data):
                    #print("    No source image to extract.")
                    continue

                try:
                    data_size = unpack(end+"I", data[:4])[0]
                    if not data_size:
                        #print('    Source data is blank.')
                        continue

                    data = bytearray(zlib.decompress(data[4:]))
                except Exception:
                    #print('    Could not decompress data.')
                    continue

                print('Extracting %s' % tag_path.split(tags_dir)[-1].lstrip("/\\"))
                try:
                    if not os.path.isdir(source_dir):
                        os.makedirs(source_dir)
                    with open(source_path, 'wb') as f:
                        a_depth = len(data) // (width*height) - 3

                        head = bytearray(18)
                        pack_into('B',  head, 2,  2)
                        pack_into('<H', head, 12, width)
                        pack_into('<H', head, 14, height)
                        pack_into('B',  head, 16, 32)
                        pack_into('B',  head, 17, 32 + ((a_depth*8)&15))
                        f.write(head)
                        f.write(data)
                except Exception:
                    #print(format_exc())
                    print("    Couldn't make Tga file.")

        print('\nFinished. Took %s seconds' % (time() - start))


if __name__ == "__main__":
    try:
        BitmapSourceExtractorWindow(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = (
    "config_def", "guerilla_workspace_def", "v2_config_def",
    "upgrade_config"
    )

from mozzarilla.defs import config_def, guerilla_workspace_def,\
     v2_config_def, upgrade_config



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

try:
    from .converter_base import ConverterBase
except (ImportError, SystemError):
    from converter_base import ConverterBase

from pathlib import Path
import threadsafe_tkinter as tk

from traceback import format_exc

from reclaimer.hek.defs.schi import schi_def
from reclaimer.hek.defs.scex import scex_def

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk


def scex_to_schi(scex_path):
    scex_tag = scex_def.build(filepath=scex_path)
    schi_tag = schi_def.build()
    schi_tag.filepath = Path(scex_path).with_suffix(".shader_transparent_chicago")

    scex_attrs = scex_tag.data.tagdata.scex_attrs
    schi_attrs = schi_tag.data.tagdata.schi_attrs

    schi_attrs.lens_flare_spacing = scex_attrs.lens_flare_spacing
    schi_attrs.extra_flags.data = scex_attrs.extra_flags.data

    schi_attrs.chicago_shader.parse(initdata=scex_attrs.chicago_shader_extended)
    schi_attrs.lens_flare.parse(initdata=scex_attrs.lens_flare)

    schi_attrs.extra_layers.STEPTREE[:] = scex_attrs.extra_layers.STEPTREE
    schi_attrs.maps.STEPTREE[:] = scex_attrs.four_stage_maps.STEPTREE

    schi_tag.calc_internal_data()

    return schi_tag


class ChicagoShaderConverter(ConverterBase, window_base_class):
    src_ext = "shader_transparent_chicago_extended"
    dst_ext = "shader_transparent_chicago"

    def __init__(self, app_root, *args, **kwargs):
        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)

        window_base_class.__init__(self, app_root, *args, **kwargs)
        ConverterBase.__init__(self, app_root, *args, **kwargs)
        kwargs["title"] = "Chicago_extended to chicago converter"
        self.setup_window(*args, **kwargs)

    def setup_window(self, *args, **kwargs):
        ConverterBase.setup_window(self, *args, **kwargs)
        self.pack_widgets()
        self.apply_style()

    def destroy(self):
        ConverterBase.destroy(self)
        window_base_class.destroy(self)

    def convert(self, tag_path):
        return scex_to_schi(tag_path)


if __name__ == "__main__":
    try:
        ChicagoShaderConverter(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

try:
    from .converter_base import ConverterBase
except (ImportError, SystemError):
    from converter_base import ConverterBase

from pathlib import Path
import threadsafe_tkinter as tk

from traceback import format_exc

from reclaimer.hek.defs.antr import antr_def
from reclaimer.os_hek.defs.magy import magy_def

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk


def magy_to_antr(magy_path):
    magy_tag = magy_def.build(filepath=magy_path)
    antr_tag = antr_def.build()
    antr_tag.filepath = Path(magy_path).with_suffix(".model_animations")

    magy_attrs = magy_tag.data.tagdata
    antr_attrs = antr_tag.data.tagdata

    antr_attrs.objects.STEPTREE[:] = magy_attrs.objects.STEPTREE
    antr_attrs.units.STEPTREE[:] = magy_attrs.units.STEPTREE
    antr_attrs.weapons.STEPTREE[:] = magy_attrs.weapons.STEPTREE
    antr_attrs.vehicles.STEPTREE[:] = magy_attrs.vehicles.STEPTREE
    antr_attrs.devices.STEPTREE[:] = magy_attrs.devices.STEPTREE
    antr_attrs.unit_damages.STEPTREE[:] = magy_attrs.unit_damages.STEPTREE
    antr_attrs.fp_animations.STEPTREE[:] = magy_attrs.fp_animations.STEPTREE
    antr_attrs.sound_references.STEPTREE[:] = magy_attrs.sound_references.STEPTREE
    antr_attrs.nodes.STEPTREE[:] = magy_attrs.nodes.STEPTREE
    antr_attrs.animations.STEPTREE[:] = magy_attrs.animations.STEPTREE

    antr_attrs.limp_body_node_radius = magy_attrs.limp_body_node_radius
    antr_attrs.flags.data = magy_attrs.flags.data

    antr_tag.calc_internal_data()

    return antr_tag


class ModelAnimationsConverter(ConverterBase, window_base_class):
    src_ext = "model_animations_yelo"
    dst_ext = "model_animations"

    def __init__(self, app_root, *args, **kwargs):
        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)

        window_base_class.__init__(self, app_root, *args, **kwargs)
        ConverterBase.__init__(self, app_root, *args, **kwargs)
        self.setup_window(*args, **kwargs)

    def setup_window(self, *args, **kwargs):
        ConverterBase.setup_window(self, *args, **kwargs)
        self.pack_widgets()
        self.apply_style()

    def destroy(self):
        ConverterBase.destroy(self)
        window_base_class.destroy(self)

    def convert(self, tag_path):
        return magy_to_antr(tag_path)


if __name__ == "__main__":
    try:
        ModelAnimationsConverter(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path
import copy
import tkinter as tk

from traceback import format_exc

from supyr_struct.buffer import get_rawdata
from supyr_struct.defs.audio.wav import wav_def

from binilla.widgets.field_widgets import FieldWidget, RawdataFrame
from binilla.windows.filedialog import askopenfilename, asksaveasfilename

from reclaimer.meta.wrappers.byteswapping import byteswap_pcm16_samples


class HaloRawdataFrame(RawdataFrame):

    def delete_node(self):
        if None in (self.parent, self.node):
            return

        undo_node = self.node
        self.node = self.parent[self.attr_index] = self.node[0:0]
        self.set_edited()

        self.edit_create(undo_node=undo_node, redo_node=self.node)

        # reload the parent field widget so sizes will be updated
        try:
            self.f_widget_parent.reload()
        except Exception:
            print(format_exc())
            print("Could not reload after deleting data.")

    def edit_apply(self=None, *, edit_state, undo=True):
        attr_index = edit_state.attr_index

        w_parent, parent = FieldWidget.get_widget_and_node(
            nodepath=edit_state.nodepath, tag_window=edit_state.tag_window)

        if undo:
            parent[attr_index] = edit_state.undo_node
        else:
            parent[attr_index] = edit_state.redo_node

        if w_parent is not None:
            try:
                w = w_parent.f_widgets[
                    w_parent.f_widget_ids_map[attr_index]]
                if w.desc is not edit_state.desc:
                    return

                w.node = parent[attr_index]
                w.set_edited()
                w.f_widget_parent.reload()
            except Exception:
                print(format_exc())


class HaloScriptSourceFrame(HaloRawdataFrame):
    @property
    def field_ext(self): return '.hsc'


class SoundSampleFrame(HaloRawdataFrame):

    @property
    def field_ext(self):
        '''The export extension of this FieldWidget.'''
        try:
            if self.parent.parent.compression.enum_name == 'ogg':
                return '.ogg'
        except Exception:
            pass
        return '.wav'

    def import_node(self):
        '''Prompts the user for an exported node file.
        Imports data into the node from the file.'''
        try:
            initialdir = self.tag_window.app_root.last_load_dir
        except AttributeError:
            initialdir = None

        ext = self.field_ext

        filepath = askopenfilename(
            initialdir=initialdir, defaultextension=ext,
            filetypes=[(self.name, "*" + ext), ('All', '*')],
            title="Import sound data from...", parent=self)

        if not filepath:
            return

        filepath = Path(filepath)
        ext = filepath.suffix.lower()

        curr_size = None
        index = self.attr_index

        try:
            curr_size = self.parent.get_size(attr_index=index)
            if ext == '.wav':
                # if the file is wav, we need to give it a header
                wav_file = wav_def.build(filepath=filepath)

                sound_data = self.parent.get_root().data.tagdata
                channel_count = sound_data.encoding.data + 1
                sample_rate = 22050 * (sound_data.sample_rate.data + 1)

                wav_header = wav_file.data.wav_header
                wav_format = wav_file.data.wav_format
                wav_chunks = wav_file.data.wav_chunks
                typ = wav_format.fmt.enum_name
                block_align = (2 if typ == "pcm" else 36) * wav_format.channels

                data_chunk = None
                for chunk in wav_chunks:
                    if chunk.sig.enum_name == "data":
                        data_chunk = chunk
                        break

                if wav_header.riff_sig != wav_header.get_desc("DEFAULT", "riff_sig"):
                    raise ValueError(
                        "RIFF signature is invalid. Not a valid wav file.")
                elif wav_header.wave_sig != wav_header.get_desc("DEFAULT", "wave_sig"):
                    raise ValueError(
                        "WAVE signature is invalid. Not a valid wav file.")
                elif wav_format.sig != wav_format.get_desc("DEFAULT", "sig"):
                    raise ValueError(
                        "Format signature is invalid. Not a valid wav file.")
                elif data_chunk is None:
                    raise ValueError(
                        "Data chunk not present. Not a valid wav file.")
                elif typ not in ('ima_adpcm', 'xbox_adpcm', 'pcm'):
                    raise TypeError(
                        "Wav file audio format must be either IMA ADPCM " +
                        "Xbox ADPCM, or PCM, not %s" % wav_format.fmt.enum_name)
                elif sound_data.encoding.data + 1 != wav_format.channels:
                    raise TypeError(
                        "Wav file channel count does not match this sound " +
                        "tags channel count. Expected %s, not %s" %
                        (channel_count, wav_format.channels))
                elif sample_rate != wav_format.sample_rate:
                    raise TypeError(
                        "Wav file sample rate does not match this sound " +
                        "tags sample rate. Expected %skHz, not %skHz" %
                        (sample_rate, wav_format.sample_rate))
                elif block_align != wav_format.block_align:
                    raise TypeError(
                        "Wav file block size does not match this sound " +
                        "tags block size. Expected %sbytes, not %sbytes" %
                        (block_align, wav_format.block_align))

                rawdata = data_chunk.data
                do_pcm_byteswap = (typ == 'pcm')
            else:
                rawdata = get_rawdata(filepath=filepath, writable=False)
                do_pcm_byteswap = False

            undo_node = self.node
            self.parent.set_size(len(rawdata), attr_index=index)
            self.parent.parse(rawdata=rawdata, attr_index=index)

            if do_pcm_byteswap:
                byteswap_pcm16_samples(self.parent)

            self.node = self.parent[index]

            self.set_edited()
            self.edit_create(undo_node=undo_node, redo_node=self.node)

            # reload the parent field widget so sizes will be updated
            try:
                self.f_widget_parent.reload()
            except Exception:
                print(format_exc())
                print("Could not reload after importing sound data.")
        except Exception:
            print(format_exc())
            print("Could not import sound data.")
            try: self.parent.set_size(curr_size, attr_index=index)
            except Exception: pass

    def export_node(self):
        try:
            initialdir = self.tag_window.app_root.last_load_dir
        except AttributeError:
            initialdir = None

        def_ext = self.field_ext

        filepath = asksaveasfilename(
            initialdir=initialdir, title="Export sound data to...",
            parent=self, filetypes=[(self.name, '*' + def_ext), ('All', '*')])

        if not filepath:
            return

        filepath = Path(filepath)
        if not filepath.suffix:
            filepath = filepath.with_suffix(def_ext)

        if filepath.suffix.lower() == '.wav':
            # if the file is wav, we need to give it a header
            try:
                wav_file = wav_def.build()
                wav_file.filepath = filepath
                sound_data = self.parent.get_root().data.tagdata

                wav_format = wav_file.data.wav_format
                wav_chunks = wav_file.data.wav_chunks
                wav_chunks.append(case="data")
                data_chunk = wav_chunks[-1]

                wav_format.bits_per_sample = 16
                wav_format.channels = sound_data.encoding.data + 1
                wav_format.sample_rate = 22050 * (sound_data.sample_rate.data + 1)

                wav_format.byte_rate = ((wav_format.sample_rate *
                                      wav_format.bits_per_sample *
                                      wav_format.channels) // 8)

                typ = "ima_adpcm"
                if self.parent.parent.compression.enum_name == 'none':
                    typ = "pcm"

                audio_data = copy.deepcopy(self.parent)
                if typ == "pcm":
                    wav_format.fmt.set_to('pcm')
                    wav_format.block_align = 2 * wav_format.channels
                    byteswap_pcm16_samples(audio_data)
                else:
                    wav_format.fmt.set_to('ima_adpcm')
                    wav_format.block_align = 36 * wav_format.channels

                data_chunk.data = audio_data.data
                wav_file.data.wav_header.filesize = wav_file.data.binsize - 12

                wav_file.serialize(temp=False, backup=False, int_test=False)
            except Exception:
                print(format_exc())
                print("Could not export sound data.")
            return

        try:
            if hasattr(self.node, 'serialize'):
                self.node.serialize(
                    filepath=filepath, clone=self.export_clone,
                    calc_pointers=self.export_calc_pointers)
            else:
                # the node isnt a block, so we need to call its parents
                # serialize method with the attr_index necessary to export.
                self.parent.serialize(
                    filepath=filepath, clone=self.export_clone,
                    calc_pointers=self.export_calc_pointers,
                    attr_index=self.attr_index)
        except Exception:
            print(format_exc())
            print("Could not export sound data.")



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from binilla.defs import upgrade_config

def upgrade_v2_to_v3(old_config, new_config):
    upgrade_config.upgrade_v1_to_v2(old_config, new_config)
    new_config.data.mozzarilla.parse(initdata=old_config.data.mozzarilla)

    return new_config



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import tkinter as tk
import time

from pathlib import Path
from tkinter import messagebox
from traceback import format_exc

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askdirectory, askopenfilename

from reclaimer.hek.defs.antr import antr_def as halo_antr_def
from reclaimer.stubbs.defs.antr import antr_def as stubbs_antr_def
from reclaimer.animation.animation_compression import \
     compress_animation, decompress_animation

from mozzarilla import editor_constants as e_c
from supyr_struct.util import is_path_empty

if __name__ == "__main__":
    window_base_class = tk.Tk
else:
    window_base_class = tk.Toplevel


class AnimationsCompressionWindow(window_base_class, BinillaWidget):
    app_root = None

    _working = False
    _loading = False

    _anims_tree_iids = ()

    def __init__(self, app_root, *args, **kwargs):
        if window_base_class == tk.Toplevel:
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)
            self.app_root = app_root
        else:
            self.app_root = self

        window_base_class.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, *args, **kwargs)

        #self.title("Model_animations compressor/decompressor")
        self.title("Model_animations decompressor")
        #self.resizable(1, 1)
        self.resizable(0, 0)
        self.update()
        anims_dir = getattr(app_root, "tags_dir", str(Path.cwd()))

        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        self.model_animations_path = tk.StringVar(self)
        self.model_animations_dir = tk.StringVar(self, anims_dir if anims_dir else "")
        self.preserve_compressed = tk.IntVar(self, True)
        self.overwrite = tk.IntVar(self, False)


        # make the frames
        self.main_frame = tk.Frame(self)
        self.anims_info_frame = tk.LabelFrame(
            self, text="Animations info")

        self.model_animations_path_frame = tk.LabelFrame(
            self.main_frame, text="Model_animations path")
        self.model_animations_path_buttons_frame = tk.Frame(self.main_frame)
        self.model_animations_dir_frame = tk.LabelFrame(
            self.main_frame, text="Model_animations dir")
        self.model_animations_dir_buttons_frame  = tk.Frame(self.main_frame)
        self.settings_frame = tk.Frame(self.main_frame)

        self.preserve_compressed_cbtn = tk.Checkbutton(
            self.settings_frame, anchor="w", variable=self.preserve_compressed,
            text="Preserve compressed animation data in tag")
        self.overwrite_cbtn = tk.Checkbutton(
            self.settings_frame, anchor="w", variable=self.overwrite,
            text="Overwrite model_animations tags")

        self.anims_info_tree = tk.ttk.Treeview(
            self.anims_info_frame, selectmode='browse', padding=(0, 0), height=4)
        self.anims_info_vsb = tk.Scrollbar(
            self.anims_info_frame, orient='vertical',
            command=self.anims_info_tree.yview)
        self.anims_info_hsb = tk.Scrollbar(
            self.anims_info_frame, orient='horizontal',
            command=self.anims_info_tree.xview)
        self.anims_info_tree.config(yscrollcommand=self.anims_info_vsb.set,
                                    xscrollcommand=self.anims_info_hsb.set)

        self.model_animations_path_entry = tk.Entry(
            self.model_animations_path_frame, width=50,
            textvariable=self.model_animations_path,
            state=tk.DISABLED)
        self.model_animations_path_browse_button = tk.Button(
            self.model_animations_path_frame, text="Browse",
            command=self.model_animations_path_browse)

        self.model_animations_dir_entry = tk.Entry(
            self.model_animations_dir_frame, width=50,
            textvariable=self.model_animations_dir,
            state=tk.DISABLED)
        self.model_animations_dir_browse_button = tk.Button(
            self.model_animations_dir_frame, text="Browse",
            command=self.model_animations_dir_browse)

        self.compress_button = tk.Button(
            self.model_animations_path_buttons_frame, text="Compress tag",
            command=self.compress_model_animations)
        self.decompress_button = tk.Button(
            self.model_animations_path_buttons_frame, text="Decompress tag",
            command=self.decompress_model_animations)

        self.compress_all_button = tk.Button(
            self.model_animations_dir_buttons_frame, text="Compress all",
            command=self.compress_all_model_animations)
        self.decompress_all_button = tk.Button(
            self.model_animations_dir_buttons_frame, text="Decompress all",
            command=self.decompress_all_model_animations)

        self.populate_animations_info_tree()

        # pack everything
        self.main_frame.pack(fill="both", side='left', pady=3, padx=3)
        #self.anims_info_frame.pack(fill="both", side='left', pady=3, padx=3, expand=True)

        self.model_animations_dir_frame.pack(fill='x')
        self.model_animations_dir_buttons_frame.pack(fill="x", pady=3, padx=3)
        self.model_animations_path_frame.pack(fill='x')
        self.model_animations_path_buttons_frame.pack(fill="x", pady=3, padx=3)
        self.settings_frame.pack(fill="both")

        for w in (self.preserve_compressed_cbtn, self.overwrite_cbtn):
            w.pack(expand=True, fill='both')

        self.model_animations_dir_entry.pack(side='left', fill='x', expand=True)
        self.model_animations_dir_browse_button.pack(side='left')

        self.model_animations_path_entry.pack(side='left', fill='x', expand=True)
        self.model_animations_path_browse_button.pack(side='left')

        self.anims_info_hsb.pack(side="bottom", fill='x')
        self.anims_info_vsb.pack(side="right",  fill='y')
        self.anims_info_tree.pack(side='left', fill='both', expand=True)

        #self.compress_all_button.pack(side='right', fill='both', padx=3, expand=True)
        self.decompress_all_button.pack(side='right', fill='both', padx=3, expand=True)
        #self.compress_button.pack(side='right', fill='both', padx=3, expand=True)
        self.decompress_button.pack(side='right', fill='both', padx=3, expand=True)

        self.apply_style()
        if self.app_root is not self:
            self.transient(self.app_root)

    def populate_animations_info_tree(self):
        anims_tree = self.anims_info_tree
        if not anims_tree['columns']:
            anims_tree['columns'] = ('data', )
            anims_tree.heading("#0")
            anims_tree.heading("data")
            anims_tree.column("#0", minwidth=100, width=100)
            anims_tree.column("data", minwidth=80, width=80, stretch=False)

        for iid in self._anims_tree_iids:
            anims_tree.delete(iid)

        self._anims_tree_iids = []

    def model_animations_dir_browse(self, force=False):
        if not force and (self._working or self._loading):
            return

        dirpath = askdirectory(
            initialdir=self.model_animations_dir.get(), parent=self,
            title="Directory of model_animations to compress/decompress")

        if not dirpath:
            return

        self.app_root.last_load_dir = dirpath
        self.model_animations_dir.set(dirpath)

    def model_animations_path_browse(self, force=False):
        if not force and (self._working or self._loading):
            return

        antr_dir = os.path.dirname(self.model_animations_path.get())
        if self.model_animations_dir.get() and not antr_dir:
            antr_dir = self.model_animations_dir.get()

        fp = askopenfilename(
            initialdir=antr_dir, title="Model_animations to compress/decompress", parent=self,
            filetypes=(("Model animations graph", "*.model_animations"), ('All', '*')))

        if not fp:
            return

        fp = Path(fp)
        if not fp.suffix:
            fp = fp.with_suffix(".model_animations")

        self.app_root.last_load_dir = str(fp.parent)
        self.model_animations_path.set(str(fp))

    def apply_style(self, seen=None):
        BinillaWidget.apply_style(self, seen)
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        window_base_class.destroy(self)

    def compress_all_model_animations(self):
        if not self._working and not self._loading:
            self._working = True
            try:
                self._do_all_compression(True)
            except Exception:
                print(format_exc())
            self._working = False

    def decompress_all_model_animations(self):
        if not self._working and not self._loading:
            self._working = True
            try:
                self._do_all_compression(False)
            except Exception:
                print(format_exc())
            self._working = False

    def compress_model_animations(self):
        if not self._working and not self._loading:
            self._working = True
            try:
                self._do_compression(True)
            except Exception:
                print(format_exc())
            self._working = False

    def decompress_model_animations(self):
        if not self._working and not self._loading:
            self._working = True
            try:
                self._do_compression(False)
            except Exception:
                print(format_exc())
            self._working = False

    def _do_all_compression(self, compress):
        antr_dir = self.model_animations_dir.get()
        while not antr_dir:
            self.model_animations_dir_browse(True)
            antr_dir = self.model_animations_dir.get()
            if not antr_dir and self.warn_cancel():
                print("    Model_animations %sion cancelled." % compress)
                return

        for root, _, files in os.walk(antr_dir):
            for fname in files:
                try:
                    if Path(fname).suffix.lower() != ".model_animations":
                        continue
                    self._do_compression(compress, Path(root, fname))
                except Exception:
                    pass#print(format_exc())

    def _do_compression(self, compress, antr_path=None):
        state = "compress" if compress else "decompress"
        if not antr_path:
            antr_path = self.model_animations_path.get()

        antr_path = Path(antr_path)

        while is_path_empty(antr_path):
            self.model_animations_path_browse(True)
            antr_path = Path(self.model_animations_path.get())
            if is_path_empty(antr_path) and self.warn_cancel():
                return

        print("%sing %s." % (state.capitalize(), antr_path))

        self.app_root.update()
        antr_def = None
        try:
            with antr_path.open('rb') as f:
                f.seek(36)
                tag_type = f.read(4)
                if tag_type == b'antr':
                    f.seek(56)
                    antr_ver = f.read(2)
                    if antr_ver == b'\x00\x05':
                        antr_def = stubbs_antr_def
                    elif antr_ver == b'\x00\x04':
                        antr_def = halo_antr_def
        except Exception:
            pass

        if antr_def is None:
            print("Could not determine model_animation tag version.")
            return

        antr_tag = antr_def.build(filepath=antr_path)
        anims = antr_tag.data.tagdata.animations.STEPTREE
        errors = False
        edited = False
        for anim in anims:
            try:
                if not anim.flags.compressed_data:
                    continue

                if compress:
                    edited |= compress_animation(anim)
                else:
                    edited |= decompress_animation(anim, self.preserve_compressed.get())
            except Exception:
                print(format_exc())
                self.update()
                errors = True

        if not edited:
            print("    No changes made. Not saving.")
            return

        if errors:
            self.update()
            if not messagebox.askyesno(
                    "Model_animations %sing failed" % state,
                    ("Errors occurred while %sing(check console). " % state) +
                     "Do you want to save the model_animations tag anyway?",
                    icon='warning', parent=self):
                print("    Model_animations compilation failed.")
                return

        try:
            if not self.overwrite.get():
                fp = Path(antr_tag.filepath)
                antr_tag.filepath = Path(
                    fp.parent, fp.stem + "_DECOMP" + fp.suffix)

            antr_tag.calc_internal_data()
            antr_tag.serialize(temp=False, backup=False, calc_pointers=False,
                               int_test=False)
            print("    Finished")
        except Exception:
            print(format_exc())
            print("    Could not save %sed model_animations." % state)

    def warn_cancel(self):
        return bool(messagebox.askyesno(
            "Unsaved model_animations",
            "Are you sure you wish to cancel?",
            icon='warning', parent=self))


if __name__ == "__main__":
    AnimationsCompressionWindow(None).mainloop()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

__all__ = (
    "SearchAndReplaceWindow", "SauceRemovalWindow",
    "BitmapSourceExtractorWindow", "BitmapConverterWindow",
    "TagScannerWindow", "DependencyWindow", "DataExtractionWindow",
    "bitmap_from_dds", "bitmap_from_multiple_dds", "bitmap_from_bitmap_source",
    "ModelCompilerWindow", "physics_from_jms",
    "hud_message_text_from_hmt", "strings_from_txt",
    "AnimationsCompilerWindow", "AnimationsCompressionWindow",
    "SoundCompilerWindow",)

from mozzarilla.windows.tools.sauce_removal_window import SauceRemovalWindow
from mozzarilla.windows.tools.dependency_window import DependencyWindow
from mozzarilla.windows.tools.data_extraction_window import DataExtractionWindow
from mozzarilla.windows.tools.search_and_replace_window import SearchAndReplaceWindow
from mozzarilla.windows.tools.bitmap_source_extractor_window import BitmapSourceExtractorWindow
from mozzarilla.windows.tools.bitmap_converter_window import BitmapConverterWindow
from mozzarilla.windows.tools.tag_scanner_window import TagScannerWindow

from mozzarilla.windows.tools.animations_compression_window import AnimationsCompressionWindow
from mozzarilla.windows.tools.animations_compiler_window import AnimationsCompilerWindow
from mozzarilla.windows.tools.model_compiler_window import ModelCompilerWindow
from mozzarilla.windows.tools.sound_compiler_window import SoundCompilerWindow

from mozzarilla.windows.tools.compile_bitmap import bitmap_from_dds, bitmap_from_multiple_dds,\
     bitmap_from_bitmap_source
from mozzarilla.windows.tools.compile_physics import physics_from_jms
from mozzarilla.windows.tools.compile_hud_message_text import hud_message_text_from_hmt
from mozzarilla.windows.tools.compile_strings import strings_from_txt



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import tkinter as tk

from binilla.widgets.binilla_widget import BinillaWidget

from mozzarilla import editor_constants as e_c


class SearchAndReplaceWindow(BinillaWidget, tk.Toplevel):
    '''
    Finds (and replaces) occurances of a string in the last touched tag window.
    '''
    app_root = None

    def __init__(self, app_root, *args, **kwargs):
        self.app_root = app_root
        kwargs.update(width=450, height=270, bd=0, highlightthickness=0)
        tk.Toplevel.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, app_root, *args, **kwargs)

        self.title("Search and Replace")
        self.minsize(width=450, height=270)
        self.resizable(1, 0)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        # make the tkinter variables
        self.find_var = tk.StringVar(self)
        self.replace_var = tk.StringVar(self)

        # make the frames
        self.comment_frame = tk.Frame(
            self, relief='sunken', bd=self.comment_depth,
            bg=self.comment_bg_color)
        self.find_frame = tk.LabelFrame(self, text="Find this")
        self.replace_frame = tk.LabelFrame(self, text="Replace with this")

        self.search_button = tk.Button(
            self, text='Count occurrances', width=20, command=self.search)
        self.replace_button = tk.Button(
            self, text='Replace occurrances', width=20, command=self.replace)

        self.find_entry = tk.Entry(
            self.find_frame, textvariable=self.find_var)
        self.replace_entry = tk.Entry(
            self.replace_frame, textvariable=self.replace_var)
        self.comment = tk.Label(
            self.comment_frame, anchor='nw', bg=self.comment_bg_color,
            justify='left', font=self.get_font("comment"),
            text="""Things to note:
  Only strings can be found/replaced. If you type in a number,
  a string consisting of that number will be searched/replaced.

  If the replacement is too long to use, you will be alerted.

  You cannot undo/redo these replacements, so be careful.""")

        self.comment.pack(side='left', fill='both', expand=True)
        self.comment_frame.pack(fill='both', expand=True)

        self.find_frame.pack(fill="x", expand=True, padx=5)
        self.find_entry.pack(fill="x", expand=True, padx=5, pady=2)
        self.search_button.pack(fill="x", anchor='center', padx=5, pady=(0,4))

        self.replace_frame.pack(fill="x", expand=True, padx=5)
        self.replace_entry.pack(fill="x", expand=True, padx=5, pady=2)
        self.replace_button.pack(fill="x", anchor='center', padx=5, pady=(0,4))

        self.transient(self.app_root)
        self.apply_style()
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        tk.Toplevel.destroy(self)

    def search(self, e=None):
        self.search_and_replace()

    def replace(self, e=None):
        self.search_and_replace(True)

    def search_and_replace(self, replace=False):
        if not self.app_root:
            print("app_root object is invalid.")
            return
        elif not self.app_root.selected_tag:
            print("No tag selected for search and replace.")
            return

        window = self.app_root.get_tag_window_by_tag(self.app_root.selected_tag)

        if window is None:
            print("Could not locate the window the selected tag is displayed by.")
            return

        find_str = self.find_var.get()
        replace_str = self.replace_var.get()

        f_widgets = window.field_widget.f_widgets.values()
        nodes = window.tag.data
        occurances = 0

        while nodes:
            new_nodes = []
            for node in nodes:
                if not isinstance(node, list):
                    continue

                attrs = range(len(node))
                if hasattr(node, 'STEPTREE'):
                    attrs = tuple(attrs) + ('STEPTREE',)
                for i in attrs:
                    val = node[i]
                    if not isinstance(val, str) or find_str != val:
                        continue

                    if not replace:
                        occurances += 1
                        continue

                    desc = node.get_desc(i)
                    f_type = desc['TYPE']

                    field_max = desc.get('MAX', f_type.max)
                    if field_max is None:
                        field_max = desc.get('SIZE')
                    replace_size = f_type.sizecalc(replace_str)
                    if replace_size > field_max:
                        print("String replacement must be less than " +
                               "%s bytes when encoded, not %s." % (
                                   field_max, replace_size))
                        continue

                    occurances += 1
                    node[i] = replace_str
                try:
                    if isinstance(node, list):
                        new_nodes.extend(node)
                    if hasattr(node, 'STEPTREE'):
                        new_nodes.append(node.STEPTREE)
                except Exception:
                    pass

            nodes = new_nodes

        if not replace:
            print('Found %s occurances' % occurances)
            return

        while f_widgets:
            new_f_widgets = []
            for w in f_widgets:
                try: new_f_widgets.extend(w.f_widgets.values())
                except Exception: pass

                if not hasattr(w, 'entry_string'):
                    continue

                try:
                    desc = w.desc
                    f_type = desc['TYPE']

                    # dont want to run this unless the nodes type is a string
                    if not isinstance(f_type.node_cls, str):
                        continue

                    field_max = w.field_max
                    if field_max is None:
                        field_max = desc.get('SIZE')

                    if f_type.sizecalc(replace_str) > field_max:
                        #print("Replacement string too long to fit.")
                        continue
                except AttributeError:
                    continue

                e_str = w.entry_string
                if find_str == e_str.get():
                    e_str.set(replace_str)

            f_widgets = new_f_widgets

        print('Found and replaced %s occurances' % occurances)
        # reload the window to display the newly entered info
        window.reload()



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path

try:
    from .converter_base import ConverterBase
except (ImportError, SystemError):
    from converter_base import ConverterBase
import threadsafe_tkinter as tk

from traceback import format_exc

from binilla.widgets.scroll_menu import ScrollMenu

from reclaimer.os_v4_hek.defs.obje import obje_def
from reclaimer.os_v4_hek.defs.bipd import bipd_def
from reclaimer.os_v4_hek.defs.vehi import vehi_def
from reclaimer.os_v4_hek.defs.weap import weap_def
from reclaimer.os_v4_hek.defs.eqip import eqip_def
from reclaimer.os_v4_hek.defs.garb import garb_def
from reclaimer.os_v4_hek.defs.proj import proj_def
from reclaimer.os_v4_hek.defs.scen import scen_def
from reclaimer.os_v4_hek.defs.mach import mach_def
from reclaimer.os_v4_hek.defs.ctrl import ctrl_def
from reclaimer.os_v4_hek.defs.lifi import lifi_def
from reclaimer.os_v4_hek.defs.plac import plac_def
from reclaimer.os_v4_hek.defs.ssce import ssce_def

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk


def get_obje_def(obje_type):
    if obje_type == "biped":
        return bipd_def
    elif obje_type == "vehicle":
        return vehi_def
    elif obje_type == "weapon":
        return weap_def
    elif obje_type == "equipment":
        return eqip_def
    elif obje_type == "garbage":
        return garb_def
    elif obje_type == "projectile":
        return proj_def
    elif obje_type == "scenery":
        return scen_def
    elif obje_type == "device_machine":
        return mach_def
    elif obje_type == "device_control":
        return ctrl_def
    elif obje_type == "device_light_fixture":
        return lifi_def
    elif obje_type == "placeholder":
        return plac_def
    elif obje_type == "sound_scenery":
        return ssce_def
    else:
        return obje_def


def obje_to_obje(obje_path, src_type, dst_type):
    src_def = get_obje_def(src_type)
    dst_def = get_obje_def(dst_type)
    if not(src_def and dst_def) or src_def is dst_def:
        return

    src_tag = src_def.build(filepath=obje_path)
    dst_tag = dst_def.build()
    dst_tag.filepath = Path(obje_path).with_suffix("." + dst_type)

    src_tagdata = src_tag.data.tagdata
    dst_tagdata = dst_tag.data.tagdata

    dst_tagdata.obje_attrs = src_tagdata.obje_attrs

    if (hasattr(dst_tagdata, "devi_attrs") and
        hasattr(src_tagdata, "devi_attrs")):
        dst_tagdata.devi_attrs = src_tagdata.devi_attrs

    if (hasattr(dst_tagdata, "unit_attrs") and
        hasattr(src_tagdata, "unit_attrs")):
        dst_tagdata.unit_attrs = src_tagdata.unit_attrs

    if (hasattr(dst_tagdata, "item_attrs") and
        hasattr(src_tagdata, "item_attrs")):
        dst_tagdata.item_attrs = src_tagdata.item_attrs

    dst_tag.calc_internal_data()

    return dst_tag


class ObjectConverter(ConverterBase, window_base_class):
    object_types = (
        "vehicle", "biped", "weapon", "equipment", "garbage",
        "device_machine", "device_control", "device_light_fixture",
        "projectile", "sound_scenery", "scenery",
        )

    def __init__(self, app_root, *args, **kwargs):
        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)

        window_base_class.__init__(self, app_root, *args, **kwargs)
        ConverterBase.__init__(self, app_root, *args, **kwargs)
        kwargs["title"] = "Object to object converter"
        self.setup_window(*args, **kwargs)

    @property
    def src_ext(self):
        try:
            return self.object_types[self.src_menu.sel_index]
        except Exception:
            return ""

    @property
    def dst_ext(self):
        try:
            return self.object_types[self.dst_menu.sel_index]
        except Exception:
            return ""

    @property
    def src_exts(self): return self.object_types

    def setup_window(self, *args, **kwargs):
        ConverterBase.setup_window(self, *args, **kwargs)

        self.settings_frame = tk.LabelFrame(self, text="Conversion settings")
        self.from_label = tk.Label(self.settings_frame, text="from")
        self.src_menu = ScrollMenu(
            self.settings_frame, options=self.object_types, menu_width=20,
            sel_index=0,
            )
        self.to_label = tk.Label(self.settings_frame, text="to")
        self.dst_menu = ScrollMenu(
            self.settings_frame, options=self.object_types, menu_width=20,
            sel_index=10,
            )

        self.pack_widgets()
        self.apply_style()

    def pack_widgets(self):
        ConverterBase.pack_widgets(self)

        # pack everything
        self.settings_frame.pack(fill='both', anchor='nw')
        self.from_label.pack(padx=(10, 0), pady=10, anchor='w', side='left')
        self.src_menu.pack(padx=(10, 0), pady=10, fill='x',
                           anchor='w', side='left')
        self.to_label.pack(padx=(10, 0), pady=10, anchor='w', side='left')
        self.dst_menu.pack(padx=(10, 0), pady=10, fill='x',
                           anchor='w', side='left')

    def destroy(self):
        ConverterBase.destroy(self)
        window_base_class.destroy(self)

    def tag_path_browse(self):
        curr_tag_path = self.tag_path.get()
        ConverterBase.tag_path_browse(self)
        if curr_tag_path == self.tag_path.get(): return

        try:
            ext = Path(curr_tag_path).suffix[1:].lower()
            self.src_menu.sel_index = self.object_types.index(ext)
        except Exception:
            pass

    def convert(self, tag_path):
        return obje_to_obje(tag_path, self.src_ext, self.dst_ext)


if __name__ == "__main__":
    try:
        ObjectConverter(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import sys

def main():
    info = sys.version_info

    if info[0] < 3 or (info[0] == 3 and info[1] < 5):
        input(
            "You must have python 3.5 or higher installed to run Mozzarilla.\n"
            "You currently have %s.%s.%s installed instead." % info[:3])
        raise SystemExit(0)

    from datetime import datetime
    from traceback import format_exc

    try:
        from mozzarilla.app_window import Mozzarilla
        main_window = Mozzarilla(debug=1)
        main_window.mainloop()

    except Exception:
        exception = format_exc()
        try:
            main_window.log_file.write('\n' + exception)
        except Exception:
            try:
                with open('startup_crash.log', 'a+') as cfile:
                    time = datetime.now().strftime("%Y-%m-%d  %H:%M:%S")
                    cfile.write("\n%s%s%s\n" % ("-"*30, time, "-"*(50-len(time))))
                    cfile.write(time + exception)
            except Exception:
                pass
        print(exception, file=sys.stderr)
        return 1;

if __name__ == "__main__":
    main()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from pathlib import Path
from traceback import format_exc

from reclaimer.strings.strings_compilation import compile_hud_message_text
from supyr_struct.util import is_path_empty
from binilla.windows.filedialog import askopenfilename


def hud_message_text_from_hmt(app, fp=None):
    load_dir = app.last_data_load_dir
    tags_dir = app.tags_dir
    data_dir = app.data_dir
    if is_path_empty(tags_dir):
        tags_dir = Path("")
    if is_path_empty(data_dir):
        data_dir = Path("")

    if is_path_empty(load_dir):
        load_dir = data_dir

    if is_path_empty(fp):
        fp = askopenfilename(
            initialdir=load_dir, parent=app,
            filetypes=(("HUD messages", "*.hmt"), ("All", "*")),
            title="Select hmt file to turn into a hud_message_text tag")

    fp = Path(fp)
    if is_path_empty(fp):
        return

    try:
        app.last_data_load_dir = fp.parent

        print("Creating hud_message_text from this hmt file:")
        print("    %s" % fp)
        with fp.open("r", encoding="utf-16-le") as f:
            hmt_string_data = f.read()
    except Exception:
        print(format_exc())
        print("    Could not load hmt file.")
        return

    try:
        rel_filepath = fp.relative_to(data_dir)
    except ValueError:
        rel_filepath = Path("hud messages")

    rel_filepath = rel_filepath.with_suffix(".hud_message_text")

    tag_path = Path("")
    if not is_path_empty(rel_filepath):
        tag_path = tags_dir.joinpath(rel_filepath)

    # make the tag window
    window = app.load_tags(
        filepaths=(tag_path, ) if tag_path.is_file() else "",
        def_id='hmt ')
    if not window:
        return

    window = window[0]
    window.is_new_tag = False
    window.tag.filepath = tag_path
    window.tag.rel_filepath = rel_filepath

    error = compile_hud_message_text(window.tag, hmt_string_data)

    # reload the window to display the newly entered info
    window.reload()
    app.update_tag_window_title(window)
    if error:
        print("    Errors occurred while compiling. " +
              "Tag will not be automatically saved.")
        window.is_new_tag = True
    elif not tag_path.is_file():
        # save the tag if it doesnt already exist
        app.save_tag()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import tkinter as tk
import tkinter.ttk as ttk

from copy import copy
from pathlib import Path, PureWindowsPath
from traceback import format_exc


from binilla import constants
from binilla.windows.filedialog import askopenfilename
from binilla.widgets.field_widgets.container_frame import ContainerFrame
from mozzarilla.widgets.field_widgets.halo_1_bitmap_display import HaloBitmapDisplayButton
from supyr_struct.util import tagpath_to_fullpath


class DependencyFrame(ContainerFrame):
    open_btn = None
    browse_btn = None
    preview_btn = None
    validate_write_trace = None
    validate_entry_str = None

    def browse_tag(self):
        '''Opens a filepicker window to aid the user in referencing their
        wished tag file.'''
        if self.node is None:
            return

        try:
            try:
                tags_dir = self.tag_window.tag.tags_dir
            except AttributeError:
                return

            init_dir = tags_dir

            try:
                init_dir = Path(tagpath_to_fullpath(
                        tags_dir, self.node.filepath,
                        extension=self.tag_window.tag.ext, force_windows=True)
                    ).parent
            except Exception:
                # This path is not valid, so use our earlier assignment instead.
                pass

            filetypes = []
            for ext in sorted(self.node.tag_class.NAME_MAP):
                if ext == 'NONE':
                    continue
                filetypes.append((ext, '*.%s' % ext))
            # TODO: Create a function that can generate globs
            # that support all types valid for this field.
            if len(filetypes) > 1:
                filetypes = (('All', '*'),) + tuple(filetypes)
            else:
                filetypes.append(('All', '*'))

            filepath = askopenfilename(
                initialdir=init_dir, filetypes=filetypes,
                title="Select a tag", parent=self)

            if not filepath:
                # askopenfilename returned nothing. Quit.
                return

            # Halo tagpaths require backslashes( \ ) as dividers.
            # We take the path we got, interpret it as a native Path, then
            # convert it to a PureWindowsPath to get the backslashes safely.
            filepath = Path(filepath)
            try:
                tag_path = filepath.relative_to(tags_dir)
            # If the path is not relative just only take the filename.
            except ValueError:
                tag_path = Path(filepath.name)

            # get file extension.
            ext = tag_path.suffix.lower()
            # Remove file extension.
            tag_path = tag_path.with_suffix('')

            orig_tag_class = copy(self.node.tag_class)
            # Try to set the tagtype to the type that we selected.
            try:
                self.node.tag_class.set_to(ext[1:])
            # If we fail, try to set the tagtype to the first one that
            # has a file that exists, otherwise, set type to NONE.
            except Exception:
                self.node.tag_class.set_to('NONE')
                for filetype in filetypes:
                    ext = filetype[1][1:]
                    if tagpath_to_fullpath(
                            tags_dir, tag_path, extension=ext) is not None:
                        self.node.tag_class.set_to(ext[1:])
                        break

            internal_tag_path = str(PureWindowsPath(tag_path)).lower()

            self.edit_create(
                attr_index=('tag_class', 'filepath'),
                redo_node=dict(
                    tag_class=self.node.tag_class, filepath=internal_tag_path),
                undo_node=dict(
                    tag_class=orig_tag_class, filepath=self.node.filepath))

            self.node.filepath = internal_tag_path
            self.reload()
            self.set_edited()
        except Exception:
            print(format_exc())

    def open_tag(self):
        '''Open the referenced tag.'''
        if self.node is None:
            return

        t_w = self.tag_window
        try:
            tag, app = t_w.tag, t_w.app_root
        except AttributeError:
            return

        cur_handler = app.handler
        new_handler = t_w.handler

        try:
            tags_dir = tag.tags_dir

            self.flush()
            if not self.node.filepath:
                return

            ext = '.' + self.node.tag_class.enum_name
            # Get full path with proper capitalization if it points to a file.
            filepath = tagpath_to_fullpath(
                tags_dir, PureWindowsPath(self.node.filepath), extension=ext)

            if filepath is None and (
            new_handler.treat_mode_as_mod2 and ext == '.model'):
                filepath = tagpath_to_fullpath(
                    tags_dir,
                    PureWindowsPath(self.node.filepath),
                    extension='.gbxmodel')

            if filepath is None:
                return

            app.set_active_handler(new_handler)
            app.load_tags(filepaths=filepath)
        except Exception:
            print(format_exc())
        finally:
            app.set_active_handler(cur_handler)

    def get_dependency_tag(self):
        if self.node is None:
            return

        t_w = self.tag_window
        try:
            tags_dir, app, handler = t_w.tag.tags_dir, t_w.app_root, t_w.handler
        except AttributeError:
            return

        try:
            self.flush()
            if not self.node.filepath:
                return

            ext = '.' + self.node.tag_class.enum_name
            # Get full path with proper capitalization if it points to a file.
            filepath = tagpath_to_fullpath(
                tags_dir, PureWindowsPath(self.node.filepath), extension=ext)

            if filepath is None and (
            handler.treat_mode_as_mod2 and ext == '.model'):
                filepath = tagpath_to_fullpath(
                    tags_dir,
                    PureWindowsPath(self.node.filepath),
                    extension='.gbxmodel')

            if filepath is None:
                return

        except Exception:
            print(format_exc())

        try:
            tag = handler.get_tag(filepath + ext)
        except Exception:
            try:
                tag = handler.build_tag(filepath=filepath)
            except Exception:
                return None

        tag.tags_dir = tags_dir  # for use by bitmap preview window
        return tag

    def bitmap_preview(self, e=None):
        '''Loads the referenced bitmap tag and displays its preview window.'''
        f = self.preview_btn.display_frame
        if f is not None and f() is not None:
            return
        try:
            tag = self.get_dependency_tag()
        except Exception:
            tag = None

        if tag is None:
            return
        self.preview_btn.change_bitmap(tag)
        self.preview_btn.show_window(None, self.tag_window.app_root)

    def validate_filepath(self, *args):
        '''Checks if the referenced tag exists, sets text color to red if
        it's invalid. Sets color to default if valid.'''
        if self.node is None:
            return

        desc = self.desc
        wid = self.f_widget_ids_map.get(desc['NAME_MAP']['filepath'])
        widget = self.f_widgets.get(wid)
        if widget is None:
            return

        try:
            tags_dir = self.tag_window.tag.tags_dir
        except AttributeError:
            return

        try:
            # Get filepath directly from the typing box
            filepath = widget.data_entry.get()
            if filepath == '':
                return

            ext = '.' + self.node.tag_class.enum_name

            # Get full path with proper capitalization if it points to a file.
            filepath = tagpath_to_fullpath(
                tags_dir, PureWindowsPath(filepath), extension=ext)

            if filepath is None and (
            self.tag_window.handler.treat_mode_as_mod2 and ext == '.model'):
                filepath = tagpath_to_fullpath(
                    tags_dir, PureWindowsPath(filepath), extension='.gbxmodel')

        except Exception:
            filepath = None
            print("Validation of a filepath failed unexpectedly.")
            print(format_exc())

        if filepath is not None:
            widget.data_entry.config(fg=self.text_normal_color)
        else:
            widget.data_entry.config(fg=self.invalid_path_color)

    def pose_fields(self):
        ContainerFrame.pose_fields(self)
        picker = self.widget_picker
        tag_window = self.tag_window

        self.browse_btn = ttk.Button(
            self, width=3, text='...', command=self.browse_tag)
        self.open_btn = ttk.Button(
            self, width=5, text='Open', command=self.open_tag)
        self.preview_btn = None
        try:
            names = self.desc[0]['NAME_MAP'].keys()
            is_bitmap_dependency = len(names) == 2 and "bitmap" in names
        except Exception:
            is_bitmap_dependency = False

        if is_bitmap_dependency:
            try:
                tags_dir = tag_window.tag.tags_dir
                app_root = tag_window.app_root
            except AttributeError:
                tags_dir = ""
                app_root = None
            self.preview_btn = HaloBitmapDisplayButton(
                self, width=7, text="Preview", command=self.bitmap_preview,
                tags_dir=tags_dir)

        padx, pady, side= self.horizontal_padx, self.horizontal_pady, 'top'
        if self.desc.get('ORIENT', 'v') in 'hH':
            side = 'left'

        for wid in self.f_widget_ids:
            w = self.f_widgets[wid]
            if w.attr_index == 0:
                # hide the tag class dropdown if not showing hidden and
                # there is only one valid tag class other than NONE
                if w.desc.get('ENTRIES', 0) <= 2 and not self.get_visible(
                        constants.VISIBILITY_HIDDEN):
                    w.pack_forget()
            elif w.attr_index == 'STEPTREE':
                # make sure the filepath has a write trace attached to it
                try:
                    self.delete_validate_trace()
                except Exception:
                    pass

                self.validate_entry_str = w.entry_string
                self.create_validate_trace()
                self.validate_filepath()

        if not hasattr(self.tag_window.tag, "tags_dir"):
            # cant do anything if the tags_dir doesnt exist
            return

        for btn in (self.browse_btn, self.open_btn, self.preview_btn):
            if btn is None: continue
            btn.pack(fill='x', side=side, anchor='nw', padx=padx)

    def create_validate_trace(self):
        self.validate_write_trace = self.write_trace(
            self.validate_entry_str, self.validate_filepath)

    def delete_validate_trace(self):
        self.validate_entry_str.trace_vdelete(
            "w", self.validate_write_trace)
        self.validate_write_trace = self.validate_entry_str = None

    def reload(self):
        '''Resupplies the nodes to the widgets which display them.'''
        try:
            f_widgets = self.f_widgets

            field_indices = range(self.desc['ENTRIES'])
            # if the node has a steptree node, include its index in the indices
            if 'STEPTREE' in self.desc:
                field_indices = tuple(field_indices) + ('STEPTREE',)

            f_widget_ids_map = self.f_widget_ids_map
            if hasattr(self, "preview_btn") and self.preview_btn:
                self.preview_btn.display_frame = None

            # if any of the descriptors are different between
            # the sub-nodes of the previous and new sub-nodes,
            # then this widget will need to be repopulated.
            sub_node = None
            for i in field_indices:
                sub_desc = self.desc[i]
                if hasattr(self.node, "__getitem__"):
                    sub_node = self.node[i]

                if hasattr(sub_node, 'desc'):
                    sub_desc = sub_node.desc

                # only display the enumerator if there are more than 2 options
                if i == 0 and sub_desc['ENTRIES'] <= 2:
                    continue

                w = f_widgets.get(f_widget_ids_map.get(i))

                # if neither would be visible, dont worry about checking it
                if not self.get_visible(sub_desc.get('VISIBLE', True)) and w is None:
                    continue

                # if the descriptors are different, gotta repopulate!
                if not hasattr(w, 'desc') or w.desc is not sub_desc:
                    self.populate()
                    return

            if self.node is not None:
                for wid in self.f_widget_ids:
                    w = f_widgets[wid]

                    w.parent, w.node = self.node, self.node[w.attr_index]
                    w.reload()

            self.validate_filepath()
        except Exception:
            print(format_exc())

    def set_disabled(self, disable=True):
        disable = disable or not self.editable
        if self.node is None and not disable:
            return

        if bool(disable) != self.disabled:
            for w in (self.open_btn, self.browse_btn, self.preview_btn):
                if w:
                    w.config(state=tk.DISABLED if disable else tk.NORMAL)

        ContainerFrame.set_disabled(self, disable)

    def apply_style(self, seen=None):
        ContainerFrame.apply_style(self, seen)
        self.validate_filepath()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import tkinter as tk
import tkinter.ttk as ttk

from traceback import format_exc

from binilla.constants import DYN_NAME_PATH, DYN_I
from binilla.widgets.field_widgets.array_frame import DynamicArrayFrame


class ReflexiveFrame(DynamicArrayFrame):
    export_all_btn = None
    import_all_btn = None

    def __init__(self, *args, **kwargs):
        DynamicArrayFrame.__init__(self, *args, **kwargs)

        btn_state = tk.DISABLED if self.disabled else tk.NORMAL

        self.import_all_btn = ttk.Button(
            self.title, width=11, text='Import all',
            command=self.import_all_nodes, state=btn_state)
        self.export_all_btn = ttk.Button(
            self.buttons, width=11, text='Export all',
            command=self.export_all_nodes, state=btn_state)

        # unpack all the buttons
        for w in (self.export_btn, self.import_btn,
                  self.shift_down_btn, self.shift_up_btn,
                  self.delete_all_btn, self.delete_btn,
                  self.duplicate_btn, self.insert_btn, self.add_btn):
            w.forget()

        # pack all the buttons(and new ones)
        # due to extra frame padding, this needs one less pixel on x
        self.import_all_btn.pack(side="right", padx=(0, 3), pady=(2, 2))

        for w in (self.export_all_btn,
                  self.shift_down_btn, self.shift_up_btn,
                  self.export_btn, self.import_btn,
                  self.delete_all_btn, self.delete_btn,
                  self.duplicate_btn, self.insert_btn, self.add_btn):
            w.pack(side="right", padx=(0, 4), pady=(2, 2))

    def set_disabled(self, disable=True):
        disable = disable or not self.editable
        if self.node is None and not disable:
            return

        if bool(disable) != self.disabled:
            new_state = tk.DISABLED if disable else tk.NORMAL
            for w in (self.export_all_btn, self.import_all_btn):
                if w:
                    w.config(state=new_state)

        DynamicArrayFrame.set_disabled(self, disable)

    def generate_dynamic_options(self, options, options_to_generate):
        node, desc = self.node, self.desc
        dyn_name_path = desc.get('DYN_NAME_PATH')

        if dyn_name_path.endswith('.filepath'):
            # if it is a dependency filepath
            for i in options_to_generate:
                name = str(node[i].get_neighbor(dyn_name_path))\
                       .split('\\')[-1].split('\n')[0]
                if name:
                    options[i] = name
            return

        for i in options_to_generate:
            name = str(node[i].get_neighbor(dyn_name_path))
            if name:
                options[i] = name.split('\n')[0]

    def set_import_all_disabled(self, disable=True):
        if disable: self.import_all_btn.config(state="disabled")
        else:       self.import_all_btn.config(state="normal")

    def set_export_all_disabled(self, disable=True):
        if disable: self.export_all_btn.config(state="disabled")
        else:       self.export_all_btn.config(state="normal")

    def export_all_nodes(self):
        try:
            w = self.f_widget_parent
        except Exception:
            return
        w.export_node()

    def import_all_nodes(self):
        try:
            w = self.f_widget_parent
        except Exception:
            return
        w.import_node()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from binilla.defs.v1_config_def import v1_general, v1_array_counts,\
     tag_window_hotkeys, padding, directory_paths, v1_app_window, \
     recent_tags, open_tags
from binilla.defs.v1_style_def import v1_widths_and_heights,\
     color, depths

from supyr_struct.defs.tag_def import TagDef
from supyr_struct.field_types import *

from mozzarilla.defs.config_def import mozz_hotkeys, mozzarilla
from mozzarilla.editor_constants import v2_mozz_color_names


__all__ = (
    "get", "v2_config_def",
    )

mozz_v2_version_info = Struct("version_info",
    UEnum32("id", ('Mozz', 'zzoM'), DEFAULT='zzoM'),
    UInt32("version", DEFAULT=2),
    )

v2_mozz_colors = Array("colors",
    SUB_STRUCT=color, SIZE="array_counts.color_count",
    MAX=len(v2_mozz_color_names), MIN=len(v2_mozz_color_names),
    NAME_MAP=v2_mozz_color_names,
    GUI_NAME="Colors"
    )

v2_config_def = TagDef("mozzarilla_v2_config",
    mozz_v2_version_info,
    v1_general,
    v1_array_counts,
    v1_app_window,
    v1_widths_and_heights,
    padding,
    depths,
    open_tags,
    recent_tags,
    directory_paths,
    v2_mozz_colors,
    mozz_hotkeys,
    tag_window_hotkeys,

    mozzarilla,
    ENDIAN='<', ext=".cfg",
    )


def get():
    return v2_config_def



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import sys
from pathlib import Path

from binilla import editor_constants as b_e_c
from binilla.editor_constants import *
from binilla.widgets.binilla_widget import BinillaWidget
from binilla.widgets.font_config import FontConfig

from supyr_struct.defs.frozen_dict import FrozenDict

v2_mozz_color_names = v1_color_names + ("active_tags_directory", )
mozz_color_names = color_names + ("active_tags_directory", )
mozz_font_names = font_names + ("font_tag_preview", )

channel_name_map   = FrozenDict(a='alpha', r='red', g='green', b='blue')
channel_offset_map = FrozenDict(a=24,      r=16,    g=8,       b=0)

TITLE_WIDTH = b_e_c.TITLE_WIDTH = 28
DEF_STRING_ENTRY_WIDTH = b_e_c.TITLE_WIDTH = 30

if b_e_c.IS_WIN:
    FONT_TAG_PREVIEW_FONT_FAMILY = "System"
    FONT_TAG_PREVIEW_FONT_SIZE   = 12
else:
    FONT_TAG_PREVIEW_FONT_FAMILY = "FreeMono"
    FONT_TAG_PREVIEW_FONT_SIZE   = 12

FONT_TAG_PREVIEW_FONT_WEIGHT = "normal"
FONT_TAG_PREVIEW_FONT_SLANT  = "roman"

BinillaWidget.font_settings.update(
    font_tag_preview=FontConfig(
        family=FONT_TAG_PREVIEW_FONT_FAMILY,
        size=FONT_TAG_PREVIEW_FONT_SIZE,
        weight=FONT_TAG_PREVIEW_FONT_WEIGHT,
        slant=FONT_TAG_PREVIEW_FONT_SLANT,
        )
    )


WORKING_DIR = Path.cwd()
MOZZLIB_DIR = Path(__file__).parent

if b_e_c.IS_WIN:
    SETTINGS_DIR = Path(WORKING_DIR, "mek_config")
else:
    SETTINGS_DIR = Path(Path.home(), ".local", "share", "mek")

MOZZ_ICON_PATH = Path(MOZZLIB_DIR, "mozzarilla.ico")
if not MOZZ_ICON_PATH.is_file():
    MOZZ_ICON_PATH = Path(MOZZLIB_DIR, "icons", "mozzarilla.ico")
if not MOZZ_ICON_PATH.is_file():
    MOZZ_ICON_PATH = Path(WORKING_DIR, "mozzarilla.ico")
if not MOZZ_ICON_PATH.is_file():
    MOZZ_ICON_PATH = Path(WORKING_DIR, "icons", "mozzarilla.ico")
if not MOZZ_ICON_PATH.is_file():
    MOZZ_ICON_PATH = ""

MOZZ_BITMAP_PATH = Path(MOZZLIB_DIR, "mozzarilla.png")
if not MOZZ_BITMAP_PATH.is_file():
    MOZZ_BITMAP_PATH = Path(MOZZLIB_DIR, "icons", "mozzarilla.png")
if not MOZZ_BITMAP_PATH.is_file():
    MOZZ_BITMAP_PATH = Path(WORKING_DIR, "mozzarilla.png")
if not MOZZ_BITMAP_PATH.is_file():
    MOZZ_BITMAP_PATH = Path(WORKING_DIR, "icons", "mozzarilla.png")
if not MOZZ_BITMAP_PATH.is_file():
    MOZZ_BITMAP_PATH = ""

IS_64BIT = sys.maxsize > 2**32

del b_e_c
del FontConfig
del Path
